/**************************************************************************
  Module:  Field.h
  Purpose: A Field contains bulk data for a mesh, and is in one-to-one
           correspondence with an Indexset within a View.  Note that
           Fields can be used as a free standing array, but it's discouraged.
 ***************************************************************************/

#ifndef __FIELD_H
#define __FIELD_H

#include "KeyBase.h"
#include "MemManage.h"
#include "MemBlock.h"
#include "VErr.h"

///
/// Data private to a Field.
/// The descendent of ValueBase providing data encapsulation specific to class
/// Field which is a descendent of the parallel KeyBase hierarchy.  This class
/// contains the bulk field data stored in the block data member.  It extends
/// ValueBase with member functions specific to field bulk data including those
/// that query the data type; allocate, fill and resize the field's bulk data;
/// and return the bulk data.
///
class FieldPrivate : public ValueBase {
public:
              FieldPrivate ( int n, DataType dtype, int depth, void **data
                           , KeyBase *creatorItem
                           );
              FieldPrivate ( void (*f_)(KeyBase *item) , int depth,
                             KeyBase *creatorItem
                           );
             ~FieldPrivate ( );

   void       advance ( int n );
   void*&     data ( );
   void*&     data ( int k );
   int        depth ( ) const;
   void       fill ( double val ) ;
   void       fill ( int val ) ;
   int        length ( ) const;
   int        length ( int k ) const;
   void       makeSpace ( );
   void       makeSpace ( void *newdata );
   void       makeSpace ( int k );
   void       release ( );
   void       release ( int k );
   void       resize ( int newLen );
   DataType   type ( ) const;

public:
   MemBlock** block;          ///< up to maxdepth of these
   int        base;           ///< current index for round robin
   int        depthPriv;      ///< current depth of round robin
   int        len;            ///< duplicate of view length
   DataType   fieldType;      ///< data IO type
   int        offset;         ///< offset into memblock (bytes)
   int        stride;         ///< stride within memblock (bytes)
   int        entitySize;     ///< size of an entity (bytes)
   int        maxDepth;       ///< max depth of round robin
   void     (*func)(KeyBase *item); ///< dynamic field
   int        funcSem ;       ///< flag for now, should be semaphore

};

///
/// Vista Items used to represent integer or real fields.
/// This class extends KeyBase with member functions specific to fields
/// including those that query the data type; allocate, fill and resize the
/// field's bulk data; and return the bulk data.
///
class Field : public KeyBase {
public:
                  Field ( const char *name, int n, DataType dtype, int depth = 1
                        , void **data = NIL, TableBase *parent = NIL
                        );
                  Field ( const char *name, void (*f_)(KeyBase *item)
                        , int depth = 1, TableBase *parent = NIL
                        );
                 ~Field ( );

   FieldPrivate&  value ( );
   int*&          Int ( );
   int*&          Int ( int n );
   double*&       Real ( );
   double*&       Real ( int n );
   int            length ( );
   int            length ( int k );
   Field*         release ( );
   Field*         release ( int n );
   Field*         makeSpace ( );
   Field*         makeSpace ( void *newdata );
   Field*         makeSpace ( int n );
   Field*         advance ( int n );
   Field*         resize ( int newLen );
   Field*         fill( double val ) ;
   Field*         fill( int val ) ;
#if 0
   Field*         addDepth ( int n );
   Field*         delDepth ( int n );
#endif
   DataType       type ( );
   int            depth ( );

private:
  // The default constructor and copy operator should not be called.
                  Field();
   Field         &operator=(const Field &) ;
};



/*****************************
 *   FieldPrivate functions
 *****************************/


inline FieldPrivate::FieldPrivate ( int n, DataType dtype, int depth,
                                    void **data, KeyBase *creatorItem )
                                  : ValueBase(creatorItem)
{  base       = 0;
   depthPriv  = depth; /* this should actually start at one.. later */
   maxDepth   = depth;
   fieldType  = dtype;
   len        = n;     /* this probably won't need to be here */
                        /* since it is also in the IndexSet of the View */
   /* later... */
   offset     = 0;
   stride     = 0;
   entitySize = 0;
   func       = NIL;         /* this is not a functor object */
   funcSem    = 1 ;

   block      = new MemBlock *[depthPriv];

   for (int i=0; i<depthPriv; ++i)
   {  int   span;
      void *temp;

      if (dtype == OTC_INT)
      {  if (data == NIL) {
            if (len != 0)
               memAllocInts(len, ((int **)&temp), NULL);
            else
               temp = NIL ;
         }
         else
            temp = data[i];
         span    = sizeof(int);
      }
      else /* dtype == OTC_REAL */
      {  if (data == NIL) {
            if (len !=0)
               memAllocDoubles(len, ((double **)&temp), NULL);
            else
               temp = NIL ;
         }
         else
            temp = data[i];
         span    = sizeof(double);
      }

      /* At the moment, this is a major hack */
      /* with many unused parameters.  C'est la vie. */
      block[i] = new MemBlock(temp, len, span);
   }
   m_valueType = FieldItem;
}

inline FieldPrivate::FieldPrivate ( void (*f_)(KeyBase *item), int depth,
                                    KeyBase *creatorItem )
                                  : ValueBase(creatorItem)
{  base        = 0;
   depthPriv   = depth; /* this should actually start at one.. later */
   maxDepth    = depth;

   /* later... */
   offset      = 0;
   stride      = 0;
   entitySize  = 0;

   func        = f_;          /* install functor */
   funcSem     = 0 ;

   fieldType   = OTC_UNKNOWN; /* this will be assigned by the functor */
   len         = 0;           /* this will be assigned by the functor */

   block       = new MemBlock *[depthPriv];

   for (int i=0; i<depthPriv; ++i)  /* functor must initialize... */
      block[i] = NIL;
   m_valueType   = FieldItem;
}

inline FieldPrivate::~FieldPrivate ( )
{
   int i ;
   for (i=0; i<depthPriv; ++i)
      block[i]->unlink() ;

   delete [] block;
}

inline void FieldPrivate::advance ( int n )
{  base = (base+depthPriv+n) % depthPriv;
}

inline void*& FieldPrivate::data ( )
{  if (funcSem == 0) {
      ++funcSem ;
      func(m_creatorItem);
      --funcSem ;
   }
   return block[base]->dataPriv;
}

inline void*& FieldPrivate::data ( int k )
{
#if 0
   if (funcSem == 0) {
      ++funcSem ;
      func(m_creatorItem); /* pass in k? */
      --funcSem ;
   }
#endif

   return block[(base+k+depthPriv)%depthPriv]->dataPriv;
}

inline void FieldPrivate::release ( )
{  if (block[base]->dataPriv != NIL)
   {  memFree(&block[base]->dataPriv, NULL); /* bad, use refcount */
      block[base]->dataPriv = NIL;
      block[base]->len      = 0;
      block[base]->maxlen   = 0;
   }
}

inline void FieldPrivate::release ( int k )
{  int idx = (base+k+depthPriv)%depthPriv;
   if (block[idx]->dataPriv != NIL)
   {  memFree(&block[idx]->dataPriv, NULL); /* bad, use refcount */
      block[idx]->dataPriv = NIL;
      block[idx]->len      = 0;
      block[idx]->maxlen   = 0;
   }
}

inline void FieldPrivate::makeSpace ( )
{  if (block[base]->dataPriv == NIL)
   {  if (fieldType == OTC_INT)
         memAllocInts(len, ((int **)&block[base]->dataPriv), NULL);
      else /* OTC_REAL */
         memAllocDoubles(len, ((double **)&block[base]->dataPriv), NULL);
      block[base]->len = len;
      block[base]->maxlen = len;
   }
}

inline void FieldPrivate::makeSpace ( void *newdata )
{  if (block[base]->dataPriv == NIL)
   {  if (fieldType == OTC_INT)
         block[base]->dataPriv = newdata;
      else /* OTC_REAL */
         block[base]->dataPriv = newdata;
      block[base]->len = len;  /* --- ASSUMED --- */
      block[base]->maxlen = len;  /* --- ASSUMED --- */
   }
}

inline void FieldPrivate::makeSpace ( int k )
{  int idx = (base+k+depthPriv)%depthPriv;
   if (block[idx]->dataPriv == NIL)
   {  if (fieldType == OTC_INT)
         memAllocInts(len, ((int **)&block[idx]->dataPriv), NULL);
      else /* OTC_REAL */
         memAllocDoubles(len, ((double **)&block[idx]->dataPriv), NULL);
      block[idx]->len    = len;
      block[idx]->maxlen = len;
   }
}

inline void FieldPrivate::resize ( int newLen )
{  int i, j;
   for (i=0; i<maxDepth; ++i)
   {  if (block[i]->dataPriv != NIL)
      {  int minLen = MIN(block[i]->len, newLen);
         if (fieldType == OTC_INT)
         {  int *newStuff;
            if (newLen > block[i]->maxlen)
            {  int *oldStuff = (int *) block[i]->dataPriv;
               memAllocInts(newLen, &newStuff, NULL);
               for (j=0; j<minLen; ++j)
                  newStuff[j] = oldStuff[j];
               memFree(&block[i]->dataPriv, NULL);
               block[i]->dataPriv = (void *) newStuff;
            }
            else
               newStuff = (int *) block[i]->dataPriv;

#if 0
            for (j=minLen; j<newLen; ++j)    /* default intitializer */
               newStuff[j] = -1;
#endif
         }
         else /* OTC_REAL */
         {
#if 0
            double myNaN = 0.0;
            *((int *) &myNaN) = 0x7ffBFFFF; /* signaling NaN */
#endif
            double *newStuff;
            if (newLen > block[i]->maxlen)
            {  double *oldStuff = (double *) block[i]->dataPriv;
               memAllocDoubles(newLen, &newStuff, NULL);
               for (j=0; j<minLen; ++j)
                  newStuff[j] = oldStuff[j];
               memFree(&block[i]->dataPriv, NULL);
               block[i]->dataPriv = (void *) newStuff;
            }
            else
               newStuff = (double *) block[i]->dataPriv;

#if 0
            for (j=minLen;j<newLen; ++j)
               newStuff[j] = myNaN;
#endif
         }
         block[i]->len = newLen;
         if (newLen > block[i]->maxlen)
            block[i]->maxlen = newLen;
      }
   }
   len = newLen;
}

inline DataType FieldPrivate::type ( ) const
{  return fieldType;
}

inline int FieldPrivate::length ( ) const
{  return block[base]->len;
}

inline void FieldPrivate::fill(double val)
{
   int i ;
   int llen = length() ;
   switch (fieldType) {
      case OTC_INT:
      {
         /* Warn: losing information */
         int filler = (int) val ;
         int *buf = (int *) data() ;
         for (i=0; i<llen; ++i)
            buf[i] = filler ;
#if DEBUG
         if (VErrGetMask() & (VERR_TRUNCATION | VERR_CONVERSION)) {
            char msg[80] ;
            sprintf(msg, "Field is int, fill type is double") ;
            VErrWarn(msg, (VERR_TRUNCATION | VERR_CONVERSION)) ;
         }
#endif
      }
      break ;

      case OTC_REAL:
      {
         double *buf = (double *) data() ;
         for (i=0; i<llen; ++i)
            buf[i] = val ;
      }
      break ;

      default: /* ignore :) */ ;
   }
}


inline void FieldPrivate::fill(int val)
{
   int i ;
   int llen = length() ;
   switch (fieldType) {
      case OTC_INT:
      {
         int *buf = (int *) data() ;
         for (i=0; i<llen; ++i)
            buf[i] = val ;
      }
      break ;

      case OTC_REAL:
      {
         double filler = (double) val ;
         double *buf = (double *) data() ;
         for (i=0; i<llen; ++i)
            buf[i] = filler ;
#if DEBUG
         if (VErrGetMask() & VERR_CONVERSION) {
            char msg[80] ;
            sprintf(msg, "Field is double, fill type is int") ;
            VErrWarn(msg, VERR_CONVERSION) ;
         }
#endif
      }
      break ;

      default: /* ignore :) */ ;
   }
}


inline int FieldPrivate::length ( int k ) const
{  return block[(base+k+depthPriv)%depthPriv]->len;
}

inline int FieldPrivate::depth ( ) const
{  return depthPriv;
}






/*****************************
 *       Field functions
 *****************************/

/* The reinterpret_casts in these constructors have to be used */
/* because static_cast isn't appropriate here.  The Item class */
/* extends the API of the KeyBase class.  Both Field and Item */
/* have a root node of Itembase.  Casting to Item implies trying */
/* to recast to a sibling class, which is unsafe in every */
/* instance except for this one. */

inline Field::Field ( const char* name, int n, DataType dtype, int depth
                    , void** data, TableBase* parent
                    )
                    : KeyBase(name, NIL, parent)
{  m_valuePtr = (ValueBase*) new FieldPrivate(n, dtype, depth, data, this);
}

inline Field::Field ( const char* name, void (*f_)(KeyBase* item),
                      int depth , TableBase *parent
                    )
                    : KeyBase(name, NIL, parent)
{  m_valuePtr = (ValueBase*) new FieldPrivate(f_, depth, this);
}

inline Field::~Field ( )
{  /* do nothing */ }

inline FieldPrivate& Field::value ( )
{  return *((FieldPrivate *)m_valuePtr);
}

inline DataType Field::type ( )
{  return value().type();
}

inline int*& Field::Int ( )
{
   /* Al suggested I wrap this in a debug...*/
#if DEBUG
   static int *iNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if ((type() != OTC_INT) && (!isNil())) {
         char msg[80] ;
         sprintf(msg, "Non-Int field %s requested as Int", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return iNull ;
      }
   }
#endif
   return (int *&) value().data();
}

inline int*& Field::Int ( int n )
{
   /* Al suggested I wrap this in a debug...*/
#if DEBUG
   static int *iNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if ((type() != OTC_INT) && (!isNil())) {
         char msg[80] ;
         sprintf(msg, "Non-Int field %s requested as Int", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return iNull ;
      }
   }
#endif
   return (int *&) value().data(n);
}

inline double*& Field::Real ( )
{
   /* Al suggested I wrap this in a debug...*/
#if DEBUG
   static double *dNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if ((type() != OTC_REAL) && (!isNil())) {
         char msg[80] ;
         sprintf(msg, "Non-Real field %s requested as Real", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return dNull ;
      }
   }
#endif
   return (double *&) value().data();
}

inline double*& Field::Real ( int n )
{
   /* Al suggested I wrap this in a debug...*/
#if DEBUG
   static double *dNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if ((type() != OTC_REAL) && (!isNil())) {
         char msg[80] ;
         sprintf(msg, "Non-Real field %s requested as Real", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return dNull ;
      }
   }
#endif
   return (double*&) value().data(n);
}

inline int Field::length ( )
{  return value().length();
}

inline int Field::length ( int k )
{  return value().length(k);
}

inline Field *Field::release ( )
{  value().release();
   return this;
}

inline Field *Field::release ( int n )
{  value().release(n);
   return this;
}

inline Field *Field::makeSpace ( )
{  if (!isNil())
      value().makeSpace();
   return this;
}

inline Field  *Field::makeSpace ( void *newdata )
{  if (!isNil())
      value().makeSpace(newdata);
   return this;
}

inline Field *Field::makeSpace ( int n )
{  if (!isNil())
      value().makeSpace(n);
   return this;
}

inline Field *Field::advance ( int n )
{  value().advance(n);
   return this;
}

inline Field *Field::resize ( int newLen )
{  if (!isNil())
      value().resize(newLen);
   return this;
}

inline Field *Field::fill( double val )
{
   value().fill(val) ;
   return this ;
}

inline Field *Field::fill( int val )
{
   value().fill(val) ;
   return this ;
}

#if 0
inline Field *Field::addDepth ( int n )
{
   /* add these in the slot after the base pointer */
   /* allocate space as the data is added */
   return this;
}

inline Field *Field::delDepth ( int n )
{
   /* delete starting from base pointer slot */
   /* positve number means delete forward */
   /* and negative number means delete behind */
   /* do not allow to drop below one item */
   return this;
}
#endif

inline int Field::depth ( )
{  return value().depth();
}

#endif /* __FIELD_H */
