
/* A group is a container for items.  Groups support set operations */
/* for creating new sets. */

#include "Group.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "VErr.h"

static const int inexactMatch = 1 ;

GroupPrivate::GroupPrivate(int maxsize, KeyBase *creatorItem) :
              ValueBase(creatorItem)
{
   if (maxsize < 1) {
      itemPtr = NIL ;
      nameKey = NIL ;
      maxlength = 0 ;
   }
   else {
      itemPtr = (KeyBase **) calloc(maxsize, sizeof(KeyBase *)) ;
      nameKey = (int *)  calloc(maxsize, sizeof(int)) ;
      maxlength = maxsize ;
   }

   length = 0 ;
   m_valueType = GroupItem ;

   return ;
}

GroupPrivate::~GroupPrivate()
{
   if (maxlength > 0) {
      free(itemPtr) ;
      free(nameKey) ;
   }

   return ;
}


static int GroupItemKey(const char *name)
{
   int itemKey ;

   /* create a quick key for checking string equality */
   /* ASCII chars are 7 bits, so we can pack them. */
   itemKey = (name[0] << 24) ;
   if (name[1]) {
      itemKey += name[1] << 17 ;
      if (name[2]) {
         itemKey += name[2] << 10 ;
         if (name[3]) {
            itemKey += name[3] << 3 ;
            if (name[4])
               itemKey += inexactMatch ; /* assume inexact */
         }
      }
   }

   return itemKey ;
}


Group *GroupPrivate::add(KeyBase *item)
{
   Group *thisGroup = (Group *) m_creatorItem ;
   /* Don't allow Nil groups to grow */
   if (!thisGroup->isNil()) {
      int i, j ;
      const char *name = item->name() ;
      int itemKey = GroupItemKey(name) ;

      /* If changed to a binary search, be careful */
      /* since multiple items may have the same nameKey */
      for (i=0; i<length && (itemKey > nameKey[i]); ++i)
         /* do nothing */ ;

      if (i != length && itemKey == nameKey[i]) {
         int keyOrder ;
     
         if (!(itemKey & inexactMatch)) {
            char msg[80];
            sprintf(msg, "Attempt to add Item named %s to Group twice", name);
            VErrFatal(msg, VERR_INTERNAL);
         }

         /* We have to potentially scan a few entries */
         for (; i<length && (itemKey == nameKey[i]) ; ++i)
            if ((keyOrder = strcmp(name, itemPtr[i]->name())) <= 0)
               break ;

         if (keyOrder == 0) {
            char msg[80];
            sprintf(msg, "Attempt to add Item named %s to Group twice", name);
            VErrFatal(msg, VERR_INTERNAL);
         }
      }

      /* itemkey < nameKey[i], or i == length */

      /* first, grow the array, if necessary */
      if (length == maxlength) {
         itemPtr = (KeyBase **)
            realloc(itemPtr, 2*maxlength*sizeof(KeyBase *)) ;
         nameKey = (int *)
            realloc(nameKey, 2*maxlength*sizeof(int)) ;
         maxlength *= 2 ; /* or += 5, etc. */
      }

      for (j=length; j != i; --j) {
         itemPtr[j] = itemPtr[j-1] ;
         nameKey[j] = nameKey[j-1] ;
      }

      itemPtr[i] = item ;
      nameKey[i] = itemKey ;
      ++length ;
   }

   return thisGroup ;
}

Group *GroupPrivate::del(KeyBase *item)
{
   int i ;

   for (i=0; i<length; ++i)
      if (itemPtr[i] == item)
         break ;

   if (i != length) {
      --length ;
      for (; i<length; ++i) {
         itemPtr[i] = itemPtr[i+1] ;
         nameKey[i] = nameKey[i+1] ;
      }
   }
   else {
      char msg[80];
      sprintf(msg, "Unable to find Item named %s. Memory corruption suspected",
              item->name());
      VErrFatal(msg, VERR_INTERNAL);
   }

   return ((Group *) m_creatorItem) ;
}

KeyBase *GroupPrivate::get(const char *name)
{
   int i ;
   static KeyBase empty("Nil") ;
   KeyBase *result = &empty ;

   int itemKey = GroupItemKey(name) ;

   for (i=0; i<length && (itemKey != nameKey[i]); ++i) 
      /* do nothing */ ;

   if (i != length && (itemKey && inexactMatch)) {
      for (; i<length && (itemKey == nameKey[i]) ; ++i) {
         if (!strcmp(name, itemPtr[i]->name())) {
            result = itemPtr[i] ;
            break ;
         }
      }
   }

   return result ;
}

/* Below are all the set operations.  We need to optimize the */
/* bejeezus out of these since they are core operations.  The best */
/* way to do this is to create special case strategies based on */
/* properties of the mesh.  ie. For union, if length(A) << length(B) */
/* we may just want to Binserach the elements of A in B as we build */
/* the union array, rather than doing a merge.  For intersection, */
/* we want to first restrict the arrays to regions that overlap, and */
/* then do the intersection operation */

Group *GroupClone(const char *name, Group* A)
{
   int i ;
   int maxlen = A->value().maxlength ;
   Group *result = new Group(name, ((maxlen > 0) ? maxlen : 5)) ;
   KeyBase **rhe = result->value().itemPtr ;
   int *rheKey = result->value().nameKey ;

   KeyBase **heAI = A->value().itemPtr ;
   int *heA = A->value().nameKey ;
   int lenA = A->value().length ;

   for (i=0; i<lenA; ++i) {
      rhe[i]    = heAI[i] ;
      rheKey[i] = heA[i] ;
   }

   result->value().length = lenA ;
   result->m_enclosingTable = A->m_enclosingTable ;

   return result ;
}

Group *GroupUnion(const char *name, Group* A, Group* B) /* A | B */
{
   Group *result ;
   int lenA = A->value().length ;
   int lenB = B->value().length ;

   if (lenA == 0 || lenB == 0) {
      if (lenA == 0 && lenB == 0) {
         result = new Group(name) ;

         if (A->m_enclosingTable == B->m_enclosingTable)
            result->m_enclosingTable = A->m_enclosingTable ;
      }
      else if (lenA == 0)
         result = GroupClone(name, B) ;
      else
         result = GroupClone(name, A) ;
   }
   else {
      result = new Group(name, lenA + lenB + 5) ;
      KeyBase **heAI = A->value().itemPtr ;
      int *heA = A->value().nameKey ;
      KeyBase **heBI = B->value().itemPtr ;
      int *heB = B->value().nameKey ;
      KeyBase **rhe = result->value().itemPtr ;
      int *rheKey = result->value().nameKey ;

      int iA = 0 ;
      int iB = 0 ;
      int ir = 0 ;

      if (A->m_enclosingTable == B->m_enclosingTable)
         result->m_enclosingTable = A->m_enclosingTable ;

      while (iA != lenA && iB != lenB) {

         int cmp = heB[iB] - heA[iA] ;

         if (cmp < 0) {
            rheKey[ir] = heB[iB] ;
            rhe[ir++] = heBI[iB++] ;
         }
         else if (cmp > 0) {
            rheKey[ir] = heA[iA] ;
            rhe[ir++] = heAI[iA++] ;
         }
         else {
            if (heB[iB] & inexactMatch) {
               cmp = strcmp(heBI[iB]->name(), heAI[iA]->name()) ;
               if (cmp == 0) {
                  rheKey[ir] = heB[iB] ;
                  rhe[ir++] = heBI[iB++] ;
                  ++iA ;
               }
               else if (cmp < 0) {
                  rheKey[ir] = heB[iB] ;
                  rhe[ir++] = heBI[iB++] ;
               }
               else {
                  rheKey[ir] = heA[iA] ;
                  rhe[ir++] = heAI[iA++] ;
               }
            }
            else {
               rheKey[ir] = heB[iB] ;
               rhe[ir++] = heBI[iB++] ;
               ++iA ;
            }
         }
      }

      while (iA != lenA) {
         rheKey[ir] = heA[iA] ;
         rhe[ir++] = heAI[iA++] ;
      }

      while (iB != lenB) {
         rheKey[ir] = heB[iB] ;
         rhe[ir++] = heBI[iB++] ;
      }

      result->value().length = ir ;
   }

   return result ;
}

Group *GroupIntersect(const char *name, Group* A, Group* B) /* A & B */
{
   Group *result ;
   int lenA = A->value().length ;
   int lenB = B->value().length ;

   if (lenA == 0 || lenB == 0) {
      result = new Group(name) ;
   }
   else {
      KeyBase **heAI ;
      int *heA = A->value().nameKey ;
      KeyBase **heBI ;
      int *heB = B->value().nameKey ;
      KeyBase **rhe ;
      int *rheKey ;

      int iA = 0 ;
      int iB = 0 ;
      int ir = 0 ;
      int minA, maxA ;
      int minB, maxB ;

      /* constrain the range */
      minA = VBinarySearchLess(heA, 0, lenA, heB[0]) ;
      maxA = VBinarySearchLess(heA, 0, lenA, heB[lenB-1]) ;

      while ((minA > 0) && (heA[minA-1] == heA[minA]))
         --minA ;

      while ((maxA < lenA-1) && (heA[maxA] == heA[maxA+1]))
         ++maxA ;

      minB = VBinarySearchLess(heB, 0, lenB, heA[0]) ;
      maxB = VBinarySearchLess(heB, 0, lenB, heA[lenA-1]) ;

      while ((minB > 0) && (heB[minB-1] == heB[minB]))
         --minB ;

      while ((maxB < lenB-1) && (heB[maxB] == heB[maxB+1]))
         ++maxB ;

      heA = &heA[minA] ;
      heAI = &A->value().itemPtr[minA] ;
      lenA = maxA - minA + 1 ;

      heB = &heB[minB] ;
      heBI = &B->value().itemPtr[minB] ;
      lenB = maxB - minB + 1 ;

      result = new Group(name, MIN(lenA, lenB) + 5) ;

      rhe = result->value().itemPtr ;
      rheKey = result->value().nameKey ;
   
      while (iA != lenA && iB != lenB) {

         int cmp = heB[iB] - heA[iA] ;

         if (cmp < 0)
            ++iB ;
         else if (cmp > 0) 
            ++iA ;
         else {
            if (heB[iB] & inexactMatch) {
               cmp = strcmp(heBI[iB]->name(), heAI[iA]->name()) ;
               if (cmp == 0) {
                  rheKey[ir] = heB[iB] ;
                  rhe[ir++] = heBI[iB++] ;
                  ++iA ;
               }
               else if (cmp < 0)
                  ++iB ;
               else 
                  ++iA ;
            }
            else {
               rheKey[ir] = heB[iB] ;
               rhe[ir++] = heBI[iB++] ;
               ++iA ;
            }
         }
      }

      result->value().length = ir ;
   }

   if (A->m_enclosingTable == B->m_enclosingTable)
      result->m_enclosingTable = A->m_enclosingTable ;


   return result ;
}

Group *GroupInverse(const char *name, Group* A)  /* ~A */
{
   Group *result ;

   if (A->value().length == 0) {
      if (A->m_enclosingTable != NIL)
         result = A->m_enclosingTable->attr("_all") ;
      else
         result  = GroupClone(name, A) ;
   }
   else {
      if (A->m_enclosingTable == NIL) {
         result = new Group(name) ; /* empty group */
      }
      else {
         int i ;
         int rpos, apos ;
         Group* all = A->m_enclosingTable->attr("_all") ;

         int lenAll = all->value().length ;
         KeyBase **heAllI = all->value().itemPtr ;
         int *heAll = all->value().nameKey ;

         int lenA = A->value().length ;
         KeyBase **heAI = A->value().itemPtr ;

         KeyBase **rhe ;
         int *rheKey ;

         result = new Group(name, lenAll - lenA + 5) ;
         rhe = result->value().itemPtr ;
         rheKey = result->value().nameKey ;

         apos = 0 ;
         rpos = 0 ;
         for (i=0; i<lenAll; ++i) {
            if (heAllI[i] != heAI[apos]) {
               rhe[rpos] = heAllI[i] ;
               rheKey[rpos] = heAll[i] ;
               ++rpos ;
            }
            else {
               if (++apos == lenA) {
                  for (++i; i<lenAll; ++i) {
                     rhe[rpos] = heAllI[i] ;
                     rheKey[rpos] = heAll[i] ;
                     ++rpos ;
                  }
                  break ;
               }
            }
         }
         result->value().length = rpos ;
         delete all ;
      }
      result->m_enclosingTable = A->m_enclosingTable ;
   }
   return result ;
}

Group *GroupSubtract(const char *name, Group* A, Group* B) /* A & ~B */
{
   /* Note that this can be tricky to implment directly */
   /* because A and B are not necessarily from the same parent */

   /* Not very efficient and assumes A has an enclosingTable... */
   Group *invGrp = GroupInverse(0, B) ;
   Group *retGrp = GroupIntersect(name, A, invGrp) ;
   delete invGrp ;

   return retGrp ;
}

Group *GroupDiff(const char *name, Group* A, Group* B) /* (A|B) - (A&B) */
{
   Group *uGrp = GroupUnion(0, A, B) ;
   Group *iGrp = GroupIntersect(0, A, B) ;
   Group *retGrp = GroupSubtract(name, uGrp, iGrp) ;
   delete uGrp ;
   delete iGrp ;

   return retGrp ;
}

