/**************************************************************************
  Module:  Group.h
  Purpose: A group is used to contain an ordered Tuple of Items.
           Groups can be used as a disaccociated parameter list that
           can be associated with a component, or can be used as
           a "array based" storage mechanism for Items.  Note that
           the Vista TableBase Attribute mechanism is heavily dependent
           on groups.
 ***************************************************************************/

#ifndef __GROUP_H
#define __GROUP_H

#include "KeyBase.h"

/*
   The assumption here is that the addition and deletion of attributes
   is infrequent.  This code needs to change if that assumption is broken.
*/


///
/// Data private to a Group.
/// The descendent of ValueBase providing data encapsulation specific to class
/// Group which is a descendent of the parallel KeyBase hierarchy.  This class
/// contains the bulk group data stored in the itemPtr data member.  It extends
/// ValueBase with member functions specific to group bulk data including those
/// that add, delete and get the bulk data values.  The nameKey and itemPtr
/// data members implement the ordered tuple nature of Group.
///
class GroupPrivate : public ValueBase {
public:
             GroupPrivate ( int maxsize, KeyBase *pItem = NIL );
            ~GroupPrivate ( );

   Group*    add ( KeyBase *item );
   Group*    del ( KeyBase *item );
   KeyBase *get ( const char *name );

public:
   KeyBase **itemPtr;  /* these are sorted by address for set operations */
   int*      nameKey ; /* This array contains a guess for the item name,
                         and is used to speed set operations on groups. */
   int       length;   
   int       maxlength;
};

/* Need an iterator for this class since it will often be scanned */
/* in its entirety */

///
/// A Group is used to hold an ordered tuples of Vista Items.
/// A Group can contain a (heterogeneous) tuple of Param, Field, Attr,
/// View, Relation and Group Items.  A Group has many uses.  For example,
/// it can be used as an abstract parameter list for a component invocation.
/// Items appearing in a tuple are sorted by name.

class Group : public KeyBase {
public:
                 Group ( const char *name, int maxsize = 5 );
                ~Group () { /* do nothing */ }

   inline GroupPrivate& value ( );

   Group        *add ( KeyBase *item );
   Group        *del ( KeyBase *item );
   KeyBase      *get ( const char *name );
   /// Number of items in the Group
   int           length ( );

   /// A Group Iterator is usually used with an Attribute query
   /// on a Table.  For example:
   /// for (Group:Iter itr(node->fields()); !itr.end(); itr.next())
   ///    printf("Nodal field %s found\n", itr.item()->name()) ;
   ///
   class Iter {
      public:
         /// A Table iterator iterates on all items in a Table.
         Iter(TableBase *t) : gi(t->attr("_all")), gidx(0) {}
         /// A Table iterator can accept a depth parameter.  If this
         /// parameter is positive, the iteration will occur over
         /// d levels of child Views, starting with the base View.
         /// A negative parameter will iterate through abs(d) levels of
         /// ancestors starting with the base View.
         Iter(TableBase *t, int d) : gi(t->attr("_all", d)), gidx(0) {}
         /// If a group is passed to a Group Iterator, then the group
         /// will be deleted when the iteration is complete.  If you
         /// don't want the group to be deleted, duplicate the group
         /// using the GroupClone() method:
         /// for (Group::Iter g(GroupClone(myGroup)); !g.end(); g.next()) ;
         ///
         Iter(Group *g) : gi(g), gidx(0) {}
         /// The destructor of a Group::Iter will delete the Group.
        ~Iter() { delete gi ; }
         /// Check for end of Group.
         inline bool end() { return gidx == gi->value().length ; }
         /// Advance to the next Item in the Group.
         inline void next() { ++gidx ; }
         /// Get the current iteration Item.
         inline KeyBase *item() { return gi->value().itemPtr[gidx] ; }
         /// Reset the iterator to start from the beginning.
         inline void reset() { gidx = 0; }

      private:
         Group *gi ; ///< Temporarily container for the group.
         int gidx ;  ///< current Item index in Group tuple.
   } ;

private:
  // The default constructor and copy operator should not be called.
                 Group();
   Group        &operator=(const Group &) ;
};

Group *GroupClone(const char *name, Group* A) ;
Group *GroupUnion(const char *name, Group* A, Group* B);
Group *GroupIntersect(const char *name, Group* A, Group* B);
Group *GroupSubtract(const char *name, Group* A, Group* B);
Group *GroupDiff(const char *name, Group* A, Group* B);
Group *GroupInverse(const char *name, Group* A) ;

inline Group *GroupClone(Group *A)
{  return GroupClone(NIL, A) ;
}

inline Group::Group ( const char *name, int maxsize ) : KeyBase(name)
{  m_valuePtr = (ValueBase*) new GroupPrivate(maxsize, this);
}

inline GroupPrivate& Group::value ( )
{  return *((GroupPrivate *)m_valuePtr);
}

inline Group *Group::add ( KeyBase *item )
{  return value().add(item);
}

inline Group *Group::del ( KeyBase *item )
{  return value().del(item);
}

inline KeyBase *Group::get ( const char *name )
{  return value().get(name);
}

inline int Group::length ( )
{   return value().length ;
}

#endif
