
#include <stdio.h>
#include "hdf5.h"
#include "ViewIO.h"

/***********************************************************************
 * Class  : HDF5Reader
 * Author(s) : Keasler
 * Purpose   : Set up a data transfer class for Vista I/O
 **********************************************************************/

///
/// A class to implement reading from the HDF5 external persistent format into
/// Vista.
/// This class implements all the pure virtual functions of class VistaTransfer.
///
class HDF5Reader : public VistaTransfer {
   hid_t context[30] ; /* max 30 levels of directories */
   int tos ;  /* top of stack */

public:
   HDF5Reader(const char *fileName, int errLevel = 0) : tos(0){
      if ((context[0] = H5Fopen(fileName, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
         setError(VIO_ERR_FOPEN) ;

      if (context[0] >= 0) {
         if (errLevel != 0)
            H5Eset_auto((H5E_auto_t) H5Eprint, (void *)stderr) ;
         else
            H5Eset_auto(NULL, NULL) ;  /* Turn off warnings */
      }
   }
   ~HDF5Reader() { H5Fclose(context[0]) ; }

   void EnterScope(const char *name)
   {
      if ((context[tos+1] = H5Gopen(context[tos], name)) < 0)
         setError(VIO_ERR_SETSCOPE) ;
      else
         ++tos ;
   }

   void ArrayReal(const char *name,
                  double *data, int length, int stride)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_DOUBLE,
                              dspace, H5P_DEFAULT) ;
      if (stride == 0 || stride == sizeof(double)) {
         if (H5Dread(dset, H5T_NATIVE_DOUBLE,
                     H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
            setError(VIO_ERR_READ) ;
         }
      }
      else /* strided */ {
         double *buf = new double[length] ;
         if (H5Dread(dset, H5T_NATIVE_DOUBLE,
                     H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) {
            setError(VIO_ERR_READ) ;
         }
         else {
            int i ;
            for (i=0; i<length; ++i) {
               *data = buf[i] ;
               data = ((double *)((char *) data + stride)) ;
            }
         }
         delete [] buf ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   void ArrayInt(const char *name,
                 int *data, int length, int stride)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_INT,
                              dspace, H5P_DEFAULT) ;
      if (stride == 0 || stride == sizeof(int)) {
         if (H5Dread(dset, H5T_NATIVE_INT,
                     H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
            setError(VIO_ERR_READ) ;
         }
      }
      else /* strided */ {
         int *buf = new int[length] ;
         if (H5Dread(dset, H5T_NATIVE_INT,
                     H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) {
            setError(VIO_ERR_READ) ;
         }
         else {
            int i ;
            for (i=0; i<length; ++i) {
               *data = buf[i] ;
               data = ((int *)((char *) data + stride)) ;
            }
         }
         delete [] buf ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   void String(const char *name, char *data, int length)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_CHAR,
                              dspace, H5P_DEFAULT) ;
      if (H5Dread(dset, H5T_NATIVE_CHAR,
                  H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
         setError(VIO_ERR_READ) ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   int Length(const char *name)
   {
      int varlen ;
      hid_t dset = H5Dopen(context[tos], name) ;
      if (dset < 0) {
         varlen = -1 ;
         setError(VIO_ERR_READ) ;
      }
      else {
         hid_t dspace = H5Dget_space(dset) ;
         varlen = H5Sget_simple_extent_npoints(dspace) ;
         H5Sclose(dspace) ;
         H5Dclose(dset) ;
      }
      return varlen ;
   }

   void ExitScope(void)
   {
      if (H5Gclose(context[tos]) < 0)
         setError(VIO_ERR_SETSCOPE) ;
      else
         --tos ;
   }
} ;

