
#include "hdf5.h"
#include "ViewIO.h"


/***********************************************************************
 * Class  : HDF5Writer
 * Author(s) : Keasler
 * Purpose   : Set up a data transfer class for Vista I/O
 **********************************************************************/

///
/// A class to implement writing to the HDF5 external persistent format from
/// Vista.
/// This class implements all the pure virtual functions of class VistaTransfer.
///
class HDF5Writer : public VistaTransfer {
   hid_t context[30] ; /* max 30 levels of directories */
   int tos ;  /* top of stack */

   static herr_t getErrNum(int n, H5E_error_t *err_desc, void *value) {
      if (n == 0) {
         *((int *)value) = ((int) err_desc->min_num) * 100 +
                           ((int) err_desc->maj_num) ;
      }

      return 0 ;
   }
public:
   HDF5Writer(const char *fileName, int fileExists = 0, int errLevel = 0) : tos(0){
      if (fileExists) {
         if ((context[0] = H5Fopen(fileName, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
            setError(VIO_ERR_FOPEN) ;
      }
      else {
         context[0] = H5Fcreate(fileName, H5F_ACC_TRUNC, H5P_DEFAULT,
                                H5P_DEFAULT) ;
         if (context[0] < 0)
            setError(VIO_ERR_FCREATE) ;
      }

      if (context[0] >= 0) {
         if (errLevel != 0)
            H5Eset_auto((H5E_auto_t) H5Eprint, (void *)stderr) ;
         else
            H5Eset_auto(NULL, NULL) ;  /* Turn off warnings */
      }
   }
   ~HDF5Writer() { H5Fclose(context[0]) ; }

   void EnterScope(const char *name)
   {
      resetError() ; /* bypass setError() on purpose */
      if ((context[tos+1] = H5Gcreate(context[tos], name, 0)) < 0) {
         int currErr ;
         H5Ewalk(H5E_WALK_UPWARD, getErrNum, (void *)&currErr) ;
         if (currErr != 3610)  /* major = 10, minor = 36 */
            setError(VIO_ERR_MKSCOPE) ;
      }
      if ((context[tos+1] < 0) && !error() &&
          ((context[tos+1] = H5Gopen(context[tos], name)) < 0)) {
         setError(VIO_ERR_SETSCOPE) ;
      }
      if (!error())
         ++tos ;
   }

   void ArrayReal(const char *name, double *data, int length, int stride)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_DOUBLE,
                              dspace, H5P_DEFAULT) ;
      if (stride == 0 || stride == sizeof(double)) {
         if (H5Dwrite(dset, H5T_NATIVE_DOUBLE,
                      H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
            setError(VIO_ERR_WRITE) ;
         }
      }
      else /* strided */ {
         int i ;
         double *buf = new double[length] ;
         double *dptr = data ;
         for (i=0; i<length; ++i) {
            buf[i] = *dptr ;
            dptr = ((double *)((char *) dptr + stride)) ;
         }
         if (H5Dwrite(dset, H5T_NATIVE_DOUBLE,
                      H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) {
            setError(VIO_ERR_WRITE) ;
         }
         delete [] buf ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   void ArrayInt(const char *name, int *data, int length, int stride)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_INT,
                              dspace, H5P_DEFAULT) ;
      if (stride == 0 || stride == sizeof(int)) {
         if (H5Dwrite(dset, H5T_NATIVE_INT,
                      H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
            setError(VIO_ERR_WRITE) ;
         }
      }
      else /* strided */ {
         int i ;
         int *buf = new int[length] ;
         int *iptr = data ;
         for (i=0; i<length; ++i) {
            buf[i] = *iptr ;
            iptr = ((int *)((char *) iptr + stride)) ;
         }
         if (H5Dwrite(dset, H5T_NATIVE_INT,
                      H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) {
            setError(VIO_ERR_WRITE) ;
         }
         delete [] buf ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   void String(const char *name, char *data, int length)
   {
      hsize_t hlen = (hsize_t) length ;
      hid_t dspace = H5Screate_simple(1, &hlen, &hlen) ;
      hid_t dset  = H5Dcreate(context[tos], name, H5T_NATIVE_CHAR,
                              dspace, H5P_DEFAULT) ;
      if (H5Dwrite(dset, H5T_NATIVE_CHAR,
                   H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) {
         setError(VIO_ERR_WRITE) ;
      }
      H5Dclose(dset) ;
      H5Sclose(dspace) ;
   }

   void ExitScope(void)
   {
      if (H5Gclose(context[tos]) < 0)
         setError(VIO_ERR_SETSCOPE) ;
      else
         --tos ;
   }
} ;

