#ifndef __INDEXSET_H
#define __INDEXSET_H

#include "Vista.h"
#include "CP.h"
#include "VErr.h"

#include <stdio.h>
#include <stdlib.h>

///
/// A map from a compact set of indices, 0-N, into another larger set of
/// indices, 0-M.
/// Used to describe the inclusion of a subset in a superset.
///
class IndexSet
{
public:

   ///  Mapping from entites in current View to entities in parent View.
   ///  If map is Nil, an identity mapping will be created.
   IndexSet ( int size, int* m=NIL, bool copy=true );
   ///  Note that a structured Indexset can represent a hyperslab of any
   /// dimension, or a traditional packed n-dimensional cube.
   IndexSet ( int dims, int *extents, int *strides, int pOffset) ;
   ~IndexSet ( );

   IndexSet  *shift ( int offset );

   ///  Returns indices of parent View entities.
   int       *map ( );

   ///  Returns number of entities in current View.
   int        length ( );

   ///  Returns the allocated length for the map (Can be larger than length).
   int        maxLength ( );

   ///  Indicates whether IndexSet is setup to function in structured mode.
   bool       isStructured ( );

   ///  Indicates whether IndexSet is setup to function in unstructured mode.
   bool       isUnstructured ( );

   ///  Create unstructured IndexSet from structured IndexSet.
   void       createUnstructuredFromStructured() ;
public:
   CP_t*      m_shared;    ///< counted pointer for array of indices.
   int*       m_is;        ///< map indices (copy of pointer stored in
                           ///< 'shared' member).
   int        m_len;       ///< map length.
   int        m_maxLen;    ///< always allocate a little extra...

   int        m_spill;     ///< overflow.  Stuff accumulates here until a
   int        m_spillType; ///< unknown, sequential, sorted, unsorted.
                           ///< reordering occurs. length+spill <= maxlength
   int        m_isType;    ///< unknown, sequential, sorted, unsorted.

   /* structured indexset */
   int        m_dims ;     ///< dimensionality of current View.
   int       *m_extents ;  ///< length along each dimension in current View.
   int       *m_strides ;  ///< stride along each dimension in parent View.
   int        m_parentCornerOffset ; ///< Index in parentView which provides
                                     ///< the base location to start traversals 
                                     ///< along each extent using the given
                                     ///< strides.
}; 

/* unstructured IndexSet */
inline IndexSet::IndexSet ( int size, int* m, bool copy )
{  int* map;
   m_len       = size;

   /* improve these when I have the time... */
   m_maxLen    = m_len;
   m_spill     = 0;
   m_isType    = 0;
   m_spillType = 0;
   m_dims      = 0;     // disable structured portion of IndexSet by default.
   m_extents   = NIL ;
   m_strides   = NIL ;
   m_parentCornerOffset = 0 ;

   if (m_len < 0)
      VErrFatal("Indexset with negative length.", VERR_INTERNAL) ;

   if ((m != NIL) && (m_len != 0) && (copy == true))
   {  memAllocInts(m_len, &map, NULL);
      for (int i=0; i<m_len; ++i)
         map[i] = m[i];
   }
   else
      map = m;

   if ((m == NIL) && (m_len != 0))
   {  memAllocInts(m_len, &map, NULL);
      for (int i=0; i<m_len; ++i)
         map[i] = i;
      copy = true;
   }

   m_shared = new CP_t(map, copy);
   m_is = map;
}

/* structured IndexSet */
inline IndexSet::IndexSet ( int dims, int *extents, int *strides,
                             int pOffset)
{
   m_dims = dims ;
   if ((m_dims > 0) && (extents != NIL) && (strides != NIL)) {
      m_strides = new int[m_dims] ;
      m_extents = new int[m_dims] ;
      m_len = 1;
      for (int i=0; i<m_dims; ++i) {
         m_strides[i] = strides[i] ;
         m_extents[i] = extents[i] ;
         m_len *= m_extents[i] ;
      }
   }
   else {
      m_extents = NIL ;
      m_strides = NIL ;
      m_len = 0 ;
   }
   m_maxLen = m_len ;
   m_parentCornerOffset = pOffset ;

   m_shared = NIL ;  /* disable unstructured component by default */
   m_is     = NIL ;
   m_spill  = 0 ;
   m_isType = 0 ;
   m_spillType = 0 ;
}

inline IndexSet::~IndexSet ( )
{  if (m_shared != NIL)
      m_shared->unlink();
   if (m_extents != NIL)
      delete [] m_extents ;
   if (m_strides != NIL)
      delete [] m_strides ;
}

inline IndexSet *IndexSet::shift ( int offset )
{
   /* This is going to change the map, so it should make a copy if needed. */
   for (int i=0; i<m_len; ++i)
      m_is[i] += offset;

   return this;
}

inline int* IndexSet::map ( )
{  return m_is;
} 

inline int IndexSet::length ( )
{  return m_len;
}

inline int IndexSet::maxLength ( )
{  return m_maxLen;
}

inline bool IndexSet::isStructured ( )
{  return (m_dims != 0) ;
}

inline bool IndexSet::isUnstructured ( )
{  return (m_is != NIL) ;
}

inline void IndexSet::createUnstructuredFromStructured()
{
   /* There's probably a better way to do this.  If you see it, */
   /* send me an email and I'll replace this algorithm with yours */
   if (m_is == NIL && m_len > 0) {
      int traverseIdx[10] ;
      int traverseOffset[10] ;
      int currLocalIdx = 0 ;
      int currParentIdx = m_parentCornerOffset ;
      int fastestIndex = m_strides[0] ;
      int fastestLen   = m_extents[0] ;

      for (int i=0; i<m_dims; ++i) {
         traverseIdx[i] = 0 ;
         traverseOffset[i] = m_parentCornerOffset ;
      }

      memAllocInts(m_len, &m_is, NULL) ;

      while (currLocalIdx != m_len) {
         for (int i=0; i<fastestLen; ++i) {
            m_is[currLocalIdx++] = currParentIdx ;
            currParentIdx += fastestIndex ;
         }
         for (int i=1; i<m_dims; ++i) {
            if (++traverseIdx[i] != m_extents[i]) {
               traverseOffset[i] += m_strides[i] ;
               currParentIdx = traverseOffset[i] ;
               for (int j=1; j<i; ++j)
                  traverseOffset[j] = currParentIdx ;
               break ;
            }
            else
               traverseIdx[i] = 0 ;
         }
      }

      m_shared = new CP_t(m_is, true);
   }
}

#endif
