/**************************************************************************
  Module:  KeyBase.c
  Purpose: These are the methods needed by Table Keys (See KeyBase.h)
 ***************************************************************************/


#include "Vista.h"
#include "KeyBase.h"
#include "Param.h"
#include "Group.h"
#include "Attr.h"
#include "VErr.h"
#include <stdio.h>

static KeyBase empty("Nil") ;

KeyBase::~KeyBase ( )
{

   if (!isNil()) {

      /* Note that TableBase::attr() can return *deletable* Nil groups. */
      /* This causes a bit of complication in the logic here. */

      if ((m_enclosingTable != NIL) && (value().m_valueType != GroupItem))
         m_enclosingTable->remove(name()) ; /* expensive */

Nil_Group_Entry:  /* In theory, a Nil group can have attributes */

      if (m_attr != NIL)
      {
         if (--m_attr->m_keyCount == 0) {
            for (Group::Iter ai(m_attr); !ai.end(); ai.next())
               if (--ai.item()->m_keyCount == 0)
                  delete ai.item() ;

            /* Note attr is auto-deleted when Group::Iter is destructed */
         }
      }

      if (--(m_valuePtr->m_valueCount) == 0)
         delete m_valuePtr ;

      if (m_keyName != NIL)
         memFree(&m_keyName, NULL);
   }
   else if ((m_valuePtr != NIL) && (m_valuePtr->m_valueType == GroupItem))
      goto Nil_Group_Entry ;
}


/* add attr to item, and to enclosingTable */
KeyBase *KeyBase::setAttr(const char *name)
{
   if (!isNil())
      return m_enclosingTable->setAttr(this, name) ;
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, int val)
{
   if (!isNil()) {
      char *dbgName = this->name() ;
      KeyBase *attrItem = new Param(dbgName, &val, 1) ;
      return m_enclosingTable->setAttr(this, name, attrItem) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, double val)
{
   if (!isNil()) {
      char *dbgName = this->name() ;
      KeyBase *attrItem = new Param(dbgName, &val, 1) ;
      return m_enclosingTable->setAttr(this, name, attrItem) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, const char *val)
{
   if (!isNil()) {
      char *dbgName = this->name() ;
      KeyBase *attrItem = new Param(dbgName, val, VStrlen(val)+1) ;
      return m_enclosingTable->setAttr(this, name, attrItem) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, int *val, int n)
{
   if (!isNil()) {
      char *dbgName = this->name() ;
      KeyBase *attrItem = new Param(dbgName, val, n) ;
      return m_enclosingTable->setAttr(this, name, attrItem) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, double *val, int n)
{
   if (!isNil()) {
      char *dbgName = this->name() ;
      KeyBase *attrItem = new Param(dbgName, val, n) ;
      return m_enclosingTable->setAttr(this, name, attrItem) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(const char *name, KeyBase *val)
{
   if (!isNil()) {
      ++(val->m_keyCount) ;
      return m_enclosingTable->setAttr(this, name, val) ;
   }
   else
      return this ;
}

KeyBase *KeyBase::setAttr(Group *otherGroup, bool shareGroup)
{
   /* Actually, we want three behaviors here:
      1) share a whole attribute group          (sharGroup == true)  (Share)
      2) share individual attributes from group (shareGruop == false)(Share)
      3) clone attributes (i.e. create snapshot of old attributes) (Copy, N/A)
   */
   if (!shareGroup || m_attr != NIL) {
      /* copy the attribute group */

      if (m_attr == NIL)  /* create an attribute group if none exists */
         m_attr = new Group(name()) ;

      /* loop over attributes */
      for (Group::Iter itr(GroupClone(otherGroup)); !itr.end(); itr.next()) {
         Attr *dupAttr = (Attr *) itr.item() ;
         ++(dupAttr->m_keyCount) ; /* increment reference count for attr */
         m_attr->add(dupAttr) ; /* add the dupAttr as a local attribute */
         m_enclosingTable->setAttrGroup(this, dupAttr->name()) ;
      }
#if DEBUG
      if (shareGroup && (VErrGetMask() & VERR_INTERNAL)) {
         char msg[80] ;
         sprintf(msg,
           "Field %s copying attributes even though sharing was requested",
           name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
#endif
   }
   else if (otherGroup != NIL) {
      m_attr = otherGroup ;  /* share attribute Group */
      ++m_attr->m_keyCount ;
      for (Group::Iter itr(GroupClone(m_attr)); !itr.end(); itr.next())
         m_enclosingTable->setAttrGroup(this, itr.item()->name()) ;
   }

   return this ;
}

Attr *KeyBase::getAttr(const char *name)
{
   if (m_attr != NIL) {
      /* Warn if Nil? */
      return (Attr *) m_attr->get(name) ;
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return (Attr *) &empty ;
   }
}

KeyBase *KeyBase::getAttrItem(const char *name)
{
   if (m_attr != NIL) {
      Attr *val = (Attr *) m_attr->get(name) ;
      if (!val->isNil())
         return val->get() ;
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTFOUND) {
            char msg[80] ;
            sprintf(msg, "Attr %s not found", name) ;
            VErrWarn(msg, VERR_NOTFOUND) ;
         }
#endif
         return &empty ;
      }
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return &empty ;
   }
}

int& KeyBase::getAttrInt( const char *name )
{
   static int empty = 0 ;

   Attr *val = getAttr(name) ;
   if (!val->isNil()) {
      Param *p = (Param *) val->get() ;
      if (!p->isNil())
         return p->Int() ;
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTFOUND) {
            char msg[80] ;
            sprintf(msg, "Attr %s not found", name) ;
            VErrWarn(msg, VERR_NOTFOUND) ;
         }
#endif
         return empty ;
      }
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return empty ;
   }
}

double& KeyBase::getAttrReal( const char *name )
{
   static double empty = 0 ;

   Attr *val = getAttr(name) ;
   if (!val->isNil()) {
      Param *p = (Param *) val->get() ;
      if (!p->isNil())
         return p->Real() ;
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTFOUND) {
            char msg[80] ;
            sprintf(msg, "Attr %s not found", name) ;
            VErrWarn(msg, VERR_NOTFOUND) ;
         }
#endif
         return empty ;
      }
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return empty ;
   }
}

char*& KeyBase::getAttrString( const char *name )
{
   static char *empty = "" ;

   Attr *val = getAttr(name) ;
   if (!val->isNil()) {
      Param *p = (Param *) val->get() ;
      if (!p->isNil())
         return p->String() ;
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTFOUND) {
            char msg[80] ;
            sprintf(msg, "Attr %s not found", name) ;
            VErrWarn(msg, VERR_NOTFOUND) ;
         }
#endif
         return empty ;
      }
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return empty ;
   }
}

KeyBase *KeyBase::delAttr( const char *name)
{
   if (m_attr != NIL) {
      Attr *val = (Attr *) m_attr->get(name) ;
      if (!val->isNil()) {
         if (m_enclosingTable != NIL)
            m_enclosingTable->delAttrGroup(this, name) ;
         m_attr->del(val) ;
         if (--val->m_keyCount == 0)
            delete val ;
         return this ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTFOUND) {
            char msg[80] ;
            sprintf(msg, "Attr %s not found", name) ;
            VErrWarn(msg, VERR_NOTFOUND) ;
         }
#endif
         return &empty ;
      }
   }
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
         char msg[80] ;
         sprintf(msg, "Attr %s not found", name) ;
         VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return &empty ;
   }
}

bool KeyBase::hasAttr(const char *name)
{
   if (m_attr != NIL) {
      Attr *val = (Attr *) m_attr->get(name) ;
      if (!val->isNil())
         return true ;
   }

   return false ;
}

void KeyBase::rename(const char *newName)
{  if (m_enclosingTable == NIL) {
      memFree(&m_keyName, NULL) ;
      m_keyName = VStrdup(newName) ;
   }
   else
      m_enclosingTable->rename(newName, name()) ;
}

void KeyBase::rename(const char *newName, int idx)
{  char tmpName[100] ;
   sprintf(tmpName, "%s%d", newName, idx) ;
   rename(tmpName) ;
}

