/**************************************************************************
  Module:  KeyBase.h
  Purpose: Provides the Key Item for the TableBase class.

           In the current implementation of Vista, the TableBase object
           uses a KeyBase as the Key part of each Hashtable entry (ValueBase
           is used as the Value part of the hash table entry).
 ***************************************************************************/

#ifndef __KEYBASE_H
#define __KEYBASE_H

#include "Vista.h"
#include "TableBase.h"
#include "ValueBase.h"
#include "VUtil.h"
#include "MemManage.h"

#include <string.h>

///
/// KeyBase is the base class for all Vista objects (called Items):
///
///    View, Relation, Field, Param, Group, Attr.
///
/// A TableBase object contains KeyBase entries.  KeyBase entries point to
/// ValueBase entries.
///
/// Note that KeyBase is designed to be used as a two-tier structure.
/// Classes that are derived from KeyBase are meant to extend
/// the KeyBase API without containing any private data.
/// This is necessary for the destructors to work right.  Private data
/// members associated with a KeyBase API should be pushed down into the
/// ValueBase object.
///
/// A KeyBase contains an attr data member which is a pointer to a Group.
/// The Group contains a vector of pointers to the Attr ibutes associated 
/// with the KeyBase Item.
///

class KeyBase {
public:
              /// Constuct a key with a given name pointing
              /// at a given value.  Supply a pointer to the
              /// parent table if the Key lives in a Table.
              KeyBase ( const char* name, ValueBase* value = NIL,
                         TableBase* pT = NIL );
   virtual   ~KeyBase ( );

   /// Add a simple attribute to the Key.
   KeyBase  *setAttr ( const char *name );
   /// Add an attribute having a string value.
   KeyBase  *setAttr ( const char *name, const char *val );
   /// Add an attribute having an integer value.
   KeyBase  *setAttr ( const char *name, int val );
   /// Add an attrinute having an integer array value.
   KeyBase  *setAttr ( const char *name, int* val, int n );
   /// Add an attribute having a real-number value.
   KeyBase  *setAttr ( const char *name, double val );
   /// Add an attrinute having a real-number array value.
   KeyBase  *setAttr ( const char *name, double *val, int n );
   /// Add an attribute that references a Param, Field, Group,
   /// Relation, View or Attr.  Note that the referenced Item
   /// can be freestanding, or it can exist in a TableBase somewhere.
   /// This feature is typically used to give a standardized name
   /// for another Item that can be variable.  For instance,
   /// elem->field("energy")->setAttr("cutoff", mesh->param("melt_energy"))
   /// allows the value of cutoff to vary.
   KeyBase  *setAttr ( const char *name, KeyBase *val );
   /// Allow different Keys to share the same set of attributes.  For example,
   /// Once attributes are created for one field, there may be several
   /// other related fields that want to share those attributes.
   /// i.e. mix->field("p")->setAttr(pure->field("p")->attr) ;
   /// if the shareGroup flag is false, then the attributes will
   /// shared as singletons rather than as a group of attributes
   /// as a whole.
   KeyBase  *setAttr ( Group *otherGroup, bool shareGroup = true) ;

   /// Get an attribute of the given name associated with this Key.
   Attr      *getAttr       ( const char *name );
   /// Get the underlying Item associated with the attribute name.
   KeyBase  *getAttrItem   ( const char *name );
   /// Get the integer value that was assigned to the attribute name.
   int       &getAttrInt    ( const char *name );
   /// Get the Real value that was assigned to the the attribute name.
   double    &getAttrReal   ( const char *name );
   /// Get the character string that was assigned to the attribute name.
   char     *&getAttrString ( const char *name );

   /// Remove an attribute from the Key.  Attribute will be dereferenced,
   /// and possibly deleted.
   KeyBase  *delAttr ( const char *name );
   /// Check the Key for the existence of a named attribute.
   bool       hasAttr ( const char *name );

   /// Get the key name.
   char*      name ( );
   /// rename the key to a new name.
   void       rename (const char *newName) ;
   /// rename the key to a new name.
   void       rename (const char *newName, int idx) ;

   /// Check to see if this Key was the first owner of the
   /// associated ValueBase.
   bool       isOriginal ( );
   /// Check to see if this Key is Nil.  A Nil key is returned
   /// when a Vista is queried for non-existent data.  Nil Items
   /// have a crippled functionality.
   bool       isNil ( );

   /// Get the ValueBase object associated with this key.
   ValueBase &value ( );

public:
   KeyBase*   m_nextKey;   ///< Keys exist as chained entries in the buckets
                           ///< of the VHashTable.
   char*      m_keyName;   ///<  Keys have a unique name
   ValueBase* m_valuePtr;  ///< Each key points at a ValueBase object,
                           ///< which contains the private data associated
                           ///< with the Key Item.
   Group*     m_attr;         ///< A Group of Attr ibutes is associated
                              ///< with each Key.
   TableBase* m_enclosingTable;  ///< This is a pointer to the Table that
                              ///< this Key lives in.  Nil if Key is a
                              ///< free floating singleton.
   bool       m_isNilObject;  ///< If true, this is an invalid yet
                              ///< well-behaved Key Item.
   int        m_keyCount ;    ///< Number of objects that share this Item.
};

inline KeyBase::KeyBase ( const char* name, ValueBase* value, TableBase* pT )
{  m_keyName     = VStrdup(name); /* rounds to sizeof Integer */
   m_valuePtr    = value;
   m_attr        = NIL;
   m_enclosingTable = pT;
   m_nextKey     = NIL;
   m_keyCount    = 1 ;
   m_isNilObject = false;
   if ((name != NIL) && (strcmp(name, "Nil") == 0))
      m_isNilObject = true;
}

inline char* KeyBase::name ( )
{  return m_keyName;
}

inline bool KeyBase::isOriginal ( )
{  return (bool)(this == m_valuePtr->m_creatorItem);
}

inline bool KeyBase::isNil ( )
{  return m_isNilObject;
}

inline ValueBase& KeyBase::value ( )
{  return *m_valuePtr ;
} 

#endif /* __KEYBASE_H */
