/**************************************************************************
  Module:  MemManage
  Purpose: Basic memory manager wrappers
 ***************************************************************************/

#include "MemManage.h"

#ifndef _WIN32
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>

typedef double real8 ;
#define MAX(a, b) (((a)>(b))? (a) : (b))

/* By default, allocate an extra 5% of space for "growing" arrays */
static real8 allocGrowthPercentage = .05 ; 
static real8 allocShrinkMultiplier = 3 ;
static int   allocShrinkCycleFreq  = 50 ;

/************************************************************************
 * Function  : MemManageInit
 * Author(s) : Rob Neely
 * Purpose   : 
 ************************************************************************/

void MemManageInit(void)
{
#if HAVE_MEMWATCH
   MemWatchInit() ;
#endif

   return ;
}


/************************************************************************
 * Function  : memAllocAlignSize
 * Author(s) : Rob Neely
 * Purpose   : 
 ************************************************************************/

int memAllocAlignSize(int insize)
{

#if WARN_ABOUT_ZERO_LENGTH_MALLOCS
   if (insize == 0)
      Warn("MemAllocAlignSize",
           "Requesting 0 length malloc") ;
#endif

#if ALIGN_MALLOCS

   int sizes[] = {32, 128, 256, 512, 1024} ;

   /* If the allocation is less than 1024 (bytes), then round it
    * up to the next thing in the table.  If it's bigger than 1024,
    * the round it to the next 1024 page limit */

   if(insize == 0)
      return 0 ;
   else if (insize < sizes[0])
      return sizes[0] ;
   else if (insize < sizes[1])
      return sizes[1] ;
   else if (insize < sizes[2])
      return sizes[2] ;
   else if (insize < sizes[3])
      return sizes[3] ;
   else if (insize < sizes[4])
      return sizes[4] ;
   else 
      return (insize + (1024 - (insize % 1024))) ;
   
#else
   
   return insize ;
   
#endif
}




/************************************************************************
 * Function  : memAllocDoubles
 * Author(s) : Rob Neely
 * Purpose   : Allows the user to do a whole bunch of mallocs with
 *             a single function call.  The first argument must be
 *             the size (in doubles, not bytes) of the malloc, followed
 *             by any number of pointers to pointers, and:
 *             THE LAST ARGUMENT MUST BE NULL OR YOU WILL TRASH THINGS
 *             VERY BADLY!!!!  
 ************************************************************************/

void memAllocDoubles(int size, ...)
{
  va_list list ;
  real8 **dptr ;
  int bytecount ;
  int count = 0 ;

  va_start(list, size) ;

  bytecount = memAllocAlignSize(size*sizeof(real8)) ;

  
  while(++count) {
    dptr = va_arg(list, real8**) ;
    if(dptr == NULL) break ;
    if(size > 0) {
      *dptr = (real8*)malloc(bytecount) ;

      if(!(*dptr)) /* error */ ;

    }
    else {
      *dptr = NULL ;
    }
  }
  
  va_end(list) ;
}






/************************************************************************
 * Function  : memAllocInts
 * Author(s) : Rob Neely
 * Purpose   : See comments for memAllocDouble
 ************************************************************************/

void memAllocInts(int size, ...)
{
  va_list list ;
  int **iptr ;
  int bytecount ;
  int count = 0 ;
  
  va_start(list, size) ;

  bytecount = memAllocAlignSize(size * sizeof(int)) ;

  while(++count) {
    iptr = va_arg(list, int**) ;
    if(iptr==NULL) break ;
    if (size > 0) {
      *iptr = (int*)malloc(bytecount) ;

      if(!(*iptr)) /* error */ ;

    }
    else {
      *iptr = NULL ;
    }
  }
  
  va_end(list) ;
}





#if BLOCK_MALLOCS
/************************************************************************
 * Function  : memAllocBlockDouble
 * Author(s) : Rob Neely
 * Purpose   : Does the same thing as memAllocDouble, but only does
 *             one malloc() system call.  To use this function successfully,
 *             one must make sure that all the arrays are freed at the same
 *             time, and that the first argument in the list is the
 *             same in memFreeBlock() function.
 ************************************************************************/

void memAllocBlockDoubles(int size, ...)
{
  va_list list ;
  real8 *tmp ;
  real8 **dptr ;
  int bytecount ;
  int count = 0 ;

  va_start(list, size) ;

  /* count the number of arguments in the vararg list */
  while(++count) {
    dptr = va_arg(list, real8**) ;
    if(dptr == NULL) break ;
  }

  /* Undo the count of the last argument (the NULL terminator) */
  count-- ;

  /* Figure out how much space to allocate */
  bytecount = memAllocAlignSize(count*size*sizeof(real8)) ;

  if(size > 0) {
     tmp = (real8*)malloc(bytecount) ;

     if(!tmp) /* error */ ;

  }

  count = 0 ;
  va_start(list, size) ;
     
  while(++count) {
     dptr = va_arg(list, real8**) ;
     if(dptr == NULL) break ;
     if (size > 0) {
        *dptr =  tmp ;
        tmp   += size ;
     }
     else {
        *dptr = NULL ;
     }
  }
  
  va_end(list) ;

}






/************************************************************************
 * Function  : memAllocBlockInts
 * Author(s) : Rob Neely
 * Purpose   : See comments for memAllocDouble
 ************************************************************************/

void memAllocBlockInts(int size, ...)
{
  va_list list ;
  int *tmp ;
  int **iptr ;
  int bytecount ;
  int count = 0 ;
  
  va_start(list, size) ;

  while(++count) {
    iptr = va_arg(list, int**) ;
    if(iptr==NULL) break ;
  }

  /* Undo the count of the last argument (the NULL terminator) */
  count-- ;

  bytecount = memAllocAlignSize(count*size*sizeof(int)) ;

  if (size > 0) {
     tmp = (int*)malloc(bytecount) ;

      if(!tmp) /* error */ ;

  }

  count = 0 ;
  va_start(list, size) ;
      
  while(++count) {
    iptr = va_arg(list, int**) ;
    if(iptr==NULL) break ;
    if (size > 0) {
      *iptr = tmp ;
      tmp   += size ;
    }
    else {
      *iptr = NULL ;
    }
  }
  
  va_end(list) ;
}

#endif




/************************************************************************
 * Function  : memReallocInt
 * Author(s) : Rob Neely
 * Purpose   :
 ************************************************************************/

void memReallocInts(int newsize, ...)
{
  va_list list ;
  int **iptr ;
  int bytecount ; 
  int count = 0 ;

  va_start(list, newsize) ;

  bytecount = memAllocAlignSize(newsize*sizeof(int)) ;

  while(++count) {
    iptr = va_arg(list, int**) ;
    if(iptr==NULL) break ;
    if(newsize > 0) {
       int *tmp = NULL ;

       if (*iptr == NULL)
          tmp = (int*)malloc(bytecount) ;
       else
          tmp = (int*)realloc(*iptr, bytecount) ;

      if(!tmp) /* error */ ;

#if NOISY_REALLOC
      if (tmp != *iptr)
         Warn("memReallocInt","realloc moved int pointer") ;
#endif
      *iptr = tmp ;
    }
    else if(*iptr != NULL) {
       free(*iptr) ;
       *iptr = NULL ;
    }
  }
  va_end(list) ;
}






/************************************************************************
 * Function  : memReallocDouble
 * Author(s) : Rob Neely
 * Purpose   :
 ************************************************************************/

void memReallocDoubles(int newsize, ...)
{
  va_list list ;
  real8 **dptr ;
  int bytecount ;
  int count = 0 ;

  va_start(list, newsize) ;

  bytecount = memAllocAlignSize(newsize*sizeof(real8)) ;
  
  while(++count) {
    dptr = va_arg(list, real8**) ;
    if(dptr==NULL) break ;
    if(newsize > 0) {
       real8 *tmp;

       /* Some machines (ASCI Red) don't like reallocing of a NULL
        * pointer, even though it's perfectly legal... */
       if (*dptr == NULL)
         tmp = (real8*)malloc (bytecount);
       else
         tmp = (real8*)realloc(*dptr, bytecount);

      if(!tmp) /* error */ ;

#if NOISY_REALLOC
      if (tmp != *dptr)
         Warn("memReallocInt","realloc moved double pointer") ;
#endif
      *dptr = tmp ;
    }
    else if (*dptr != NULL) {
       free(*dptr) ;
       *dptr = NULL ;
    }
  }
  va_end(list) ;
}





/************************************************************************
 * Function  : memFree
 * Author(s) : Jim Reus (based on work by Rob Neely).
 * Purpose   : Free's a bunch of memory which was malloc'ed by
 *             the other vararg routines in this file.
 *
 *             MAKE THE LAST PARAMETER NULL!!!!
 *
 ************************************************************************/

void memFree ( void* ptr, ... )
{  va_list list;
   void**  p;

   va_start(list, ptr);
   while ((p=(void**)ptr) != NULL)
   {  if (*p != NULL)
      {  free(*p);
	 *p = NULL;
      }
#if DEBUG
      else
         Warn("memFree","Freeing NULL pointer");
#endif
      ptr = va_arg(list,void*);
   }
   va_end(list);
}




#if BLOCK_MALLOCS
/************************************************************************
 * Function  : memBlockFree
 * Author(s) : Jim Reus (based on work by Rob Neely).
 * Purpose   : Free's memory that was allocated with one of the
 *             block mallocs. Note that only the first pointer is
 *             really released (since the block allocate purchased
 *             one block to cover all the pointers).
 *
 *             MAKE THE LAST PARAMETER NULL!!!!
 *
 ************************************************************************/

void memBlockFree ( void *ptr, ...)
{  int       i;
   va_list   list;
   void    **p;

   va_start(list,ptr);
   for (i=0,p=(void**)ptr; p != NULL; ++i)
   {  if (*p != NULL)
      {  if (i == 0)
	    free(*p);
	 *p = NULL;
      }
#if DEBUG
      else
	 Warn("memBlockFree","Freeing NULL pointer") ;
#endif
      p = va_arg(list,void**);
   }
   va_end(list);
}
#endif


/************************************************************************
 * Function  : memSetAllocGrowthPct
 * Author(s) : RN
 * Purpose   : For arrays which typically grow in the code (material data,
 *             history vars, etc...) - this is the percentage of extra space
 *             which will be allocated when we need to realloc space.
 ************************************************************************/

void memSetAllocGrowthPct(real8 growPct)
{
   allocGrowthPercentage = growPct/100.0 ;
}
void memSetShrinkFreq(int cycs)
{
   allocShrinkCycleFreq = cycs ;
}
void memSetShrinkMult(real8 shrinkMult)
{
   allocShrinkMultiplier = shrinkMult ;
}

real8 memGetAllocGrowthPct(void)
{
   return allocGrowthPercentage*100.0 ;
}
int memGetShrinkFreq(void)
{
   return allocShrinkCycleFreq ;
}
real8 memGetShrinkMult(void)
{
   return allocShrinkMultiplier ;
}





/************************************************************************
 * Function  : memCalcAllocGrowth
 * Author(s) : RN
 * Purpose   : Given a requested size of a "growing" array, this function
 *             uses the supplied percentage to allocate a little extra.
 ************************************************************************/

int memCalcAllocGrowth(int size)
{
   int extraSize, allocSize ;

   extraSize = (int)(size*allocGrowthPercentage) ;
   
   /* Make sure it grows by at least 1 (unless the percentage is really
    * zero - in which case they probably really mean zero) */
   if ((extraSize == 0) && (allocGrowthPercentage > 0.0))
      extraSize = 1 ;
   
   allocSize = size + extraSize ;
   return allocSize ;
}


/************************************************************************
 * Function  : memAllocShrink
 * Author(s) : RN
 * Purpose   : Every so many cycles, looks at the amount of space allocated
 *             for the material arrays compared to the amount of space
 *             actually used - and if it's different by some multiple
 *             of the "growth" percentage we use for growing them - then
 *             it returns the new size we should allocate to.
 *
 *             Returns -1 if nothing should be done.
 ************************************************************************/

int memCalcAllocShrink(int currCycle, int actual,
                       int actualLast, int alloc)
{
   int retVal = -1 ;

   /* Input:
    *
    * currCycle - The current cycle number.  Used to determine if we
    *             should even bother to check...
    * actual    - The actual size currently required for this region.  Does
    *             not include the extra percentage tacked on.
    * actualLast - The actual size required during the previous cycle.
    *             We can't shrink to below this lest we don't leave enough
    *             space for the "old" arrays.
    * alloc     - The actual amount of space currently allocated
    *
    */

   /* Only bother to shrink if we're at a cycle which evenly divides into
    * the check frequency */
   if ((currCycle % allocShrinkCycleFreq == 0) && (actual != 0)) {
      int request = MAX(actual, actualLast) ;
      int reqAlloc = memCalcAllocGrowth(request) ;
      int reqExtra    = reqAlloc - request ;
      int actualExtra = alloc - reqAlloc ;

      if (reqExtra > 0) {
         int extraRatio = (int)((real8)actualExtra / (real8)reqExtra) ;
      
         if (extraRatio >= allocShrinkMultiplier)
            retVal = request ;
         else
            retVal = -1 ;
      }
      else {
         retVal = -1 ;
      }
   }
   return retVal ;
}

