/**************************************************************************
  Module:  Param.h
  Purpose: Parameters are Items that contain data that should be associated
           with a View scope, but that should not be in one-to-one
           correspondence with a View's Indexset.  Note that a Parameter
           can be used as a free standing data type, but that usage is
           discouraged.
 ***************************************************************************/

#ifndef __PARAM_H
#define __PARAM_H

#include "KeyBase.h"
#include "VErr.h"

///
/// Data private to a Param.
/// The descendent of ValueBase providing data encapsulation specific to class
/// Param which is a descendent of the parallel KeyBase hierarchy.  This class
/// contains the param data stored in the dataPriv data member.  It extends
/// ValueBase with member functions specific to param bulk data including those
/// that query the data type and return the parameter value.
///
class ParamPrivate : public ValueBase {
public:
            ParamPrivate ( int* value, int n,
                           bool copyMe, KeyBase *creatorItem );
            ParamPrivate ( double* value, int n,
                           bool copyMe, KeyBase *creatorItem );
            ParamPrivate ( const char* value, int n,
                           bool copyMe, KeyBase *creatorItem );
           ~ParamPrivate ( );

   void*&   data ( );
   int      length ( );
   DataType type ( );

public:
   void*    dataPriv;     /* raw data storage */
   int      len;          /* length of data type */
   DataType paramType;    /* IO data type */
   bool  priv;         /* is this data private (ie to be free())? */

private:
  // The default constructor should not be called.
            ParamPrivate();
};

///
/// Vista Items used to represent data that should be associated with a View
/// scope, but that should not be in one-to-one correspondence with a View's
/// IndexSet.
/// This class extends KeyBase with member functions specific to paramters
/// including those that query the data type and return the parameter value.
/// Note that a Param can be used as a free standing data type, but that usage
/// is discouraged.
///
class Param : public KeyBase {
public:
                 Param ( const char* name, const char* value, int n,
                         bool copyMe = VISTA_OWNS );
                 Param ( const char* name, int* value, int n,
                         bool copyMe = VISTA_OWNS );
                 Param ( const char* name, double* value, int n,
                         bool copyMe = VISTA_OWNS );
                ~Param ( );

   ParamPrivate& value ( );

   /* the following functions should probably be type checked... */
   char*&        String ( );
   int&          Int ( );
   double&       Real ( );
   int*&         IntArray ( );
   double*&      RealArray ( );
   int           length ( );
   DataType      type ( );

private:
  // The default constructor and copy operator should not be called.
                 Param();
   Param        &operator=(const Param &) ;
};


inline ParamPrivate::ParamPrivate ( int* value, int n,
                                    bool copyMe, KeyBase *creatorItem ) :
                                  ValueBase(creatorItem) 
{  int* buf;

   if ((copyMe == true) && (value != NIL))
   {  buf = new int[n];
      for (int i=0; i<n; ++i)
         buf[i] = value[i];
      priv = true;
   }
   else
   {  buf = value;
      priv = false;
   }

   dataPriv  = (void*) buf;
   len       = n;
   paramType = OTC_INT;
   m_valueType = ParamItem;
}

inline ParamPrivate::ParamPrivate ( double* value, int n,
                                    bool copyMe, KeyBase *creatorItem ) :
                                    ValueBase(creatorItem) 
{  double* buf;

   if ((copyMe == true) && (value != NIL))
   {  buf = new double[n];
      for (int i=0; i<n; ++i)
         buf[i] = value[i];
      priv = true;
   }
   else
   {  buf = value;
      priv = false;
   }

   dataPriv  = (void*) buf;
   len       = n;
   paramType = OTC_REAL;
   m_valueType = ParamItem;
}

inline ParamPrivate::ParamPrivate ( const char* value, int n,
                                    bool copyMe, KeyBase *creatorItem ) :
                                    ValueBase(creatorItem)
{  char* buf;

   if ((copyMe == true) && (value != NIL))
   {  buf = new char[n];
      for (int i=0; i<n; ++i)
         buf[i] = value[i];
      priv = true;
   }
   else
   {  buf  = (char*) value;
      priv = false;
   }

   dataPriv  = (void*) buf;
   len       = n;
   paramType = OTC_STRING;
   m_valueType = ParamItem;
}

inline ParamPrivate::~ParamPrivate()
{  if (priv)
   {
      switch (paramType)
      {
         case OTC_STRING:
            delete [] reinterpret_cast<char*>(dataPriv);
            break;

         case OTC_INT:
            delete [] reinterpret_cast<int*>(dataPriv);
            break;

         case OTC_REAL:
            delete [] reinterpret_cast<double*>(dataPriv);
            break;

         case OTC_UNKNOWN:
            // The constructors for this class only allow for string, int and
            // real parameter types.
            break;
      }
   }
}

inline void*& ParamPrivate::data ( )
{  return (void*&)dataPriv;
}

inline int ParamPrivate::length ( )
{  return len;
}

inline DataType ParamPrivate::type ( )
{  return paramType;
}

inline Param::Param ( const char* name, const char* value, int n,
                      bool copyMe ) :
                      KeyBase(name)
{  m_valuePtr = (ValueBase*) new ParamPrivate(value, n, copyMe, this);
}

inline Param::Param ( const char* name, int* value, int n,
                      bool copyMe ) :
                      KeyBase(name)
{  m_valuePtr = (ValueBase*) new ParamPrivate(value, n, copyMe, this);
}

inline Param::Param ( const char* name, double* value, int n,
                      bool copyMe ) :
                      KeyBase(name)
{  m_valuePtr = (ValueBase*) new ParamPrivate(value, n, copyMe, this);
}

inline Param::~Param ( )
{  /* do nothing */ }

inline ParamPrivate& Param::value ( )
{  return *((ParamPrivate*)m_valuePtr);
} 

inline DataType Param::type ( )
{  return value().type();
}

inline char*& Param::String ( )
{
   /* Al suggested I wrap this in a debug... */
#if DEBUG
   static char *sNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if (type() != OTC_STRING) {
         char msg[80] ;
         sprintf(msg, "Non-String param %s requested as String", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return sNull ;
      }
   }
#endif
   return (char*&)value().data();
}

inline int& Param::Int ( )
{
   /* Al suggested I wrap this in a debug... */
#if DEBUG
   static int iNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if (type() != OTC_INT) {
         char msg[80] ;
         sprintf(msg, "Non-Int param %s requested as Int", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return iNull ;
      }
   }
#endif
   return *((int*&)value().data());
}

inline double& Param::Real ( )
{
   /* Al suggested I wrap this in a debug... */
#if DEBUG
   static double dNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if (type() != OTC_REAL) {
         char msg[80] ;
         sprintf(msg, "Non-Real param %s is requested as Real", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return dNull ;
      }
   }
#endif
   return *((double*&)value().data());
}

inline int*& Param::IntArray ( )
{
   /* Al suggested I wrap this in a debug... */
#if DEBUG
   static int *iNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if (type() != OTC_INT) {
         char msg[80] ;
         sprintf(msg, "Non-Int param %s requested as Int", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return iNull ;
      }
   }
#endif
   return (int*&) value().data();
}

inline double*& Param::RealArray ( )
{
   /* Al suggested I wrap this in a debug... */
#if DEBUG
   static double *dNull = 0 ; /* too bad we can't make this const... */
   if (VErrGetMask() & VERR_CONVERSION) {
      if (type() != OTC_REAL) {
         char msg[80] ;
         sprintf(msg, "Non-Real param %s is requested as Real", name()) ;
         VErrFatal(msg, VERR_CONVERSION) ;
         return dNull ;
      }
   }
#endif
   return (double*&) value().data();
}

inline int Param::length ( )
{  return value().length();
}

#endif /* __PARAM_H */
