
#include <string.h>

#include "ParamTable.h"
#include "Param.h"
#include "View.h"
#include "Vista.h"
#include "VErr.h"

static const char *nothing = NIL ;
static Param empty("Nil", nothing, 0) ;

Param *ParamTable::param(const char *name)
{
   View *nView ;
   const char *nName ;
   VisPath(m_enclosingView, name, &nView, &nName) ;
   if (nView->isNil() || nName[0] == '\0') {
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
          char msg[80] ;
          sprintf(msg, "Param %s not found", name) ;
          VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return &empty ;
   }
   else {            /* search this View and ancestors for parameter */
      for(TableBase *pt = nView->params(); pt != NIL; pt = pt->m_parentTable) {
         if (pt->size() != 0) {
            /* fast backdoor lookup */
            Param *pp = (Param *) VHashTableFind(pt->m_itemTable, nName) ;
            if (!pp->isNil())
               return pp ;
         }
      }
#if DEBUG
      if (VErrGetMask() & VERR_NOTFOUND) {
          char msg[80] ;
          sprintf(msg, "Param %s not found", name) ;
          VErrWarn(msg, VERR_NOTFOUND) ;
      }
#endif
      return &empty ;  /* root of tree was reached with no match */
   }
}

Param *ParamTable::setInt(const char *name, int value)
{
   Param *pe = (Param *) get(name) ;

   if (pe->isNil()) {
      View *nView ;
      const char *nName ;
      VisPath(m_enclosingView, name, &nView, &nName) ;
      if (!nView->isNil() && nName[0] != '\0') {

         /* since we pass by value, Vista must own */
         Param *entry = new Param(nName, &value, 1, VISTA_OWNS) ;
         nView->params()->add(entry) ;

         /* assume first value added is default value */
         setAttr(entry, "_init",
                 (new Param(nName, &value, 1, VISTA_OWNS))) ;

         return entry ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTCREATED) {
             char msg[80] ;
             sprintf(msg, "Could not create param %s", name) ;
             VErrWarn(msg, VERR_NOTCREATED) ;
         }
#endif
         return &empty ;
      }
   }
   else {
      if (VErrGetMask() & VERR_CONVERSION) {
         if (pe->type() != OTC_INT) {
            char msg[80] ;
            sprintf(msg, "Non integer param %s not rewritten int", name) ;
            VErrFatal(msg, VERR_CONVERSION) ;
            return pe ;
         }
      }
      if (pe->type() == OTC_INT)
         pe->Int() = value ;

      return pe ;
   }
}

Param *ParamTable::setReal(const char *name, double value)
{
   Param *pe = (Param *) get(name) ;

   if (pe->isNil()) {
      View *nView ;
      const char *nName ;
      VisPath(m_enclosingView, name, &nView, &nName) ;
      if (!nView->isNil() && nName[0] != '\0') {

         /* since we pass by value, Vista must own */
         Param *entry = new Param(nName, &value, 1, VISTA_OWNS) ;
         nView->params()->add(entry) ;

         /* assume first value added is default value */
         setAttr(entry, "_init",
                 (new Param(nName, &value, 1, VISTA_OWNS))) ;

         return entry ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTCREATED) {
             char msg[80] ;
             sprintf(msg, "Could not create param %s", name) ;
             VErrWarn(msg, VERR_NOTCREATED) ;
         }
#endif
         return &empty ;
      }
   }
   else {
      if (VErrGetMask() & VERR_CONVERSION) {
         if (pe->type() != OTC_REAL) {
            char msg[80] ;
            sprintf(msg, "Non real param %s not rewritten as type real", name) ;
            VErrFatal(msg, VERR_CONVERSION) ;
            return pe ;
         }
      }
      if (pe->type() == OTC_REAL)
         pe->Real() = value ;

      return pe ;
   }
}

Param *ParamTable::setString(const char *name, const char *value,
                              int slen, bool copyMe)
{
   if (slen == 0)
      slen = ((value == NIL) ? 0 : strlen(value) + 1) ;

   Param *pe = (Param *) get(name) ;

   if (pe->isNil()) {
      View *nView ;
      const char *nName ;
      VisPath(m_enclosingView, name, &nView, &nName) ;
      if (!nView->isNil() && nName[0] != '\0') {

         Param *entry = new Param(nName, value, slen, copyMe) ;
         nView->params()->add(entry) ;

         /* assume first value added is default value */
         /* Note that this could be alot of data, so we don't necessarily */
         /* want to keep a default value.  For now, keep a default if it's */
         /* less than 1K of data */
         if (slen < 1024) {
            setAttr(entry, "_init",
                    (new Param(nName, value, slen, VISTA_OWNS))) ;
         }

         return entry ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTCREATED) {
             char msg[80] ;
             sprintf(msg, "Could not create param %s", name) ;
             VErrWarn(msg, VERR_NOTCREATED) ;
         }
#endif
         return &empty ;
      }
   }
   else {
      if (VErrGetMask() & VERR_CONVERSION) {
         if (pe->type() != OTC_STRING) {
            char msg[80] ;
            sprintf(msg, "non String param %s not rewritten as string", name) ;
            VErrFatal(msg, VERR_CONVERSION) ;
            return pe ;
         }
      }

      if (pe->type() == OTC_STRING) {
         /* mostly stolen from Param constructor... */

         char *buf ;

         /* free old data if we owned it */
         if (pe->value().priv && pe->value().len != 0) {
            /* This doesn't pay attention to reference count (bad)... */
            delete [] ((char *)pe->value().dataPriv) ;
         }

         if (copyMe == true && value != NIL)
         {  buf = new char[slen];
            for (int i=0; i<slen; ++i)
               buf[i] = value[i];
            pe->value().priv = true;
         }
         else
         {  buf  = (char*) value;
            pe->value().priv = false;
         }

         pe->value().dataPriv = (void*) buf;
         pe->value().len = slen ;
      }

      return pe ;
   }

}

Param *ParamTable::setIntArray(const char *name, int *value, int len, bool copyMe)
{
   Param* pe = (Param *) get(name) ;

   if (pe->isNil()) {
      View *nView ;
      const char *nName ;
      VisPath(m_enclosingView, name, &nView, &nName) ;
      if (!nView->isNil() && nName[0] != '\0') {

         Param *entry = new Param(nName, value, len, copyMe) ;
         nView->params()->add(entry) ;

         /* assume first value added is default value */
         /* Note that this could be alot of data, so we don't necessarily */
         /* want to keep a default value.  For now, keep a default if it's */
         /* less than 1K of data */
         if (len < 1024) {
            setAttr(entry, "_init",
                    (new Param(nName, value, len, VISTA_OWNS))) ;
         }

         return entry ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTCREATED) {
             char msg[80] ;
             sprintf(msg, "Could not create param %s", name) ;
             VErrWarn(msg, VERR_NOTCREATED) ;
         }
#endif
         return &empty ;
      }
   }
   else {
      if (VErrGetMask() & VERR_CONVERSION) {
         if (pe->type() != OTC_INT) {
            char msg[80] ;
            sprintf(msg, "Non integer param %s not rewritten as int", name) ;
            VErrFatal(msg, VERR_CONVERSION) ;
            return pe ;
         }
      }

      if (pe->type() == OTC_INT) {
         /* mostly stolen from Param constructor... */

         int *buf ;

         /* free old data if we owned it */
         if (pe->value().priv && pe->value().len != 0) {
            /* This doesn't pay attention to reference count (bad)... */
            delete [] ((int *)pe->value().dataPriv) ;
         }

         if (copyMe == true && value != NIL)
         {  buf = new int[len];
            for (int i=0; i<len; ++i)
               buf[i] = value[i];
            pe->value().priv = true;
         }
         else
         {  buf = value;
            pe->value().priv = false;
         }

         pe->value().dataPriv = (void*) buf;
         pe->value().len = len ;
      }

      return pe ;
   }
}

Param *ParamTable::setRealArray(const char *name, double *value, int len,
                                 bool copyMe)
{
   Param *pe = (Param *) get(name) ;

   if (pe->isNil()) {
      View *nView ;
      const char *nName ;
      VisPath(m_enclosingView, name, &nView, &nName) ;
      if (!nView->isNil() && nName[0] != '\0') {

         Param *entry = new Param(nName, value, len, copyMe) ;
         nView->params()->add(entry) ;

         /* assume first value added is default value */
         /* Note that this could be alot of data, so we don't necessarily */
         /* want to keep a default value.  For now, keep a default if it's */
         /* less than 1K of data */
         if (len < 1024) {
            setAttr(entry, "_init",
                    (new Param(nName, value, len, VISTA_OWNS))) ;
         }

         return entry ;
      }
      else {
#if DEBUG
         if (VErrGetMask() & VERR_NOTCREATED) {
             char msg[80] ;
             sprintf(msg, "Could not create param %s", name) ;
             VErrWarn(msg, VERR_NOTCREATED) ;
         }
#endif
         return &empty ;
      }
   }
   else {
      if (VErrGetMask() & VERR_CONVERSION) {
         if (pe->type() != OTC_REAL) {
            char msg[80] ;
            sprintf(msg, "Non real param %s not rewritten as type real", name) ;
            VErrFatal(msg, VERR_CONVERSION) ;
            return pe ;
         }
      }

      if (pe->type() == OTC_REAL) {
         /* mostly stolen from Param constructor... */

         double *buf ;

         /* free old data if we owned it */
         if (pe->value().priv && pe->value().len != 0) {
            /* This doesn't pay attention to reference count (bad)... */
            delete [] ((double *)pe->value().dataPriv) ;
         }

         if (copyMe == true && value != NIL)
         {  buf = new double[len];
            for (int i=0; i<len; ++i)
               buf[i] = value[i];
            pe->value().priv = true;
         }
         else
         {  buf = value;
            pe->value().priv = false;
         }

         pe->value().dataPriv = (void*) buf;
         pe->value().len = len ;
      }

      return pe ;
   }
}
