/**************************************************************************
  Module:  Relation.h
  Purpose: A relation provides information about topological connectivity 
           bewteen two Views.  Each entity in the originating View
           (the View the relation lives in), can be connected to one or
           more entities in another View.  For example, in a hexehedral
           mesh, an elemToNode relation created in the the hex-element
           View could connect each hex-element to eight nodes in the nodal
           View. Note that a relation can connect a View to itself, such as
           when an elemToElement relation is made to connect each element
           in a View to its neighboring elements in that View.
 ***************************************************************************/

#ifndef __RELATION_H
#define __RELATION_H

#include "KeyBase.h"

///
/// Data private to a Relation.
/// The descendent of ValueBase providing data encapsulation specific to class
/// Relation which is a descendent of the parallel KeyBase hierarchy.  This
/// class contains the bulk relation data stored in the dataPriv data member.
///
class RelationPrivate : public ValueBase
{
public:
         RelationPrivate ( int mlen, int fixedSize, View* oView,
                           KeyBase *creatorItem );
        ~RelationPrivate ( );

   int*  data ( int n );
   int   length ( );
public:
   /* sharing and all that other stuff should probably come later... */
   int*  dataPriv;
   int   size;
   int   len;
   View* otherView;
};

///
/// Vista Items used to provide information about topological connectivity 
/// bewteen two Views.
/// Each entity in the originating View (the View the relation lives in), can
/// be connected to one or more entities in another View.  For example, in a
/// hexehedral mesh, an elemToNode Relation created in the the hex-element View
/// could connect each hex-element to eight nodes in the nodal View.  Note that
/// a Relation can connect a View to itself, such as when an elemToElement
/// Relation is made to connect each element in a View to its neighboring
/// elements in that View.
/// This class extends KeyBase with member functions specific to Relations
/// including a [] operator so that this class may be used like a conventional
/// connectivity array.
///
class Relation : public KeyBase
{
public:
                    /// The mlen parameter is the number of entities in
                    /// the current View.  The fixedSize parameter
                    /// is the number of entites in the other View
                    /// that are connected to each entity in the current
                    /// View.  The oView parameter is a pointer to the
                    /// View we are mapping to.
                    Relation ( const char* name, int mlen, int fixedSize
                             , View* oView = NIL
                             );
                   ~Relation ( );

   RelationPrivate& value ( );

   /// The parameter in the brackets should be the index of the
   /// entity in the current View.  This allows the user to append
   /// another set of brackets to get the specific entity, i.e.
   /// elemToNode[48][0] ... elemToNode[48][7] gives the indices
   /// for the eight nodes in the other View that are connected to
   /// element 48 in the current View.
   int*             operator[] ( int n );
   /// This returns the mlen parameter passed into the constructor.
   int              length ( );

private:
  // The default constructor and copy operator should not be called.
                    Relation();
   Relation        &operator=(const Relation &) ;
};

inline RelationPrivate::RelationPrivate ( int mlen, int fixedSize
                                        , View* oView, KeyBase *creatorItem
                                        ) : ValueBase(creatorItem)
                                          , size(fixedSize)
                                          , len(mlen)
                                          , otherView(oView)
{  m_valueType = RelationItem;
   dataPriv  = new int[mlen*fixedSize];
}

inline RelationPrivate::~RelationPrivate ( )
{  delete dataPriv ;
}

inline int* RelationPrivate::data ( int n )
{  return &dataPriv[n*size];
}

inline int RelationPrivate::length ( )
{  return len;
}

inline Relation::Relation ( const char* name, int mlen, int fixedSize, View* oView
                          ) : KeyBase(name)
{  m_valuePtr = (ValueBase*) new RelationPrivate(mlen, fixedSize, oView, this);
}

inline Relation::~Relation ( )
{  /* do nothing */ }

inline RelationPrivate& Relation::value ( )
{  return *((RelationPrivate*)m_valuePtr);
}

inline int* Relation::operator[] ( int n )
{  return value().data(n);
}

inline int Relation::length ( )
{  return value().length();
}

#endif /* __RELATION_H */
