
#include "silo.h"
#include "ViewIO.h"

/***********************************************************************
 * Class  : SILOWriter
 * Author(s) : Keasler
 * Purpose   : Set up a data transfer class for Vista I/O
 **********************************************************************/

///
/// A class to implement writing to the Silo external persistent format from
/// Vista.
/// This class implements all the pure virtual functions of class VistaTransfer.
///
class SILOWriter : public VistaTransfer {
   DBfile *file ;

public:
   SILOWriter(const char *fileName, int fileExists = 0, int errLevel = 0)
   {
      if (fileExists) {
         if ((file = DBOpen( fileName, DB_PDB, DB_APPEND)) == NULL)
            setError(VIO_ERR_FOPEN) ;
      }
      else {
         if ((file = DBCreate( fileName, DB_CLOBBER, DB_LOCAL,
                              NULL, DB_PDB)) == NULL)
            setError(VIO_ERR_FCREATE) ;
      }

      if (file != NULL)
         DBShowErrors(((errLevel == 0) ? DB_NONE : DB_ALL), NULL) ;
   }

   ~SILOWriter() { DBClose(file) ; }

   void EnterScope(const char *name)
   {
      resetError() ; /* bypass setError() on purpose */
      if (DBMkDir(file, const_cast<char *> (name)) == -1) {
         if (db_errno != 8)
            setError(VIO_ERR_MKSCOPE) ;
      }
      if (!error() && (DBSetDir(file, const_cast<char *> (name)) == -1))
            setError(VIO_ERR_SETSCOPE) ;
   }

   void ArrayReal(const char *name, double *data, int length, int stride)
   {
      if (stride == 0 || stride == sizeof(double)) {
         if (DBWrite(file, const_cast<char *> (name),
                     data, &length, 1, DB_DOUBLE) == -1) {
            setError(VIO_ERR_WRITE) ;
         }
      }
      else /* strided */ {
         int i ;
         double *buf = new double[length] ;
         double *dptr = data ;
         for (i=0; i<length; ++i) {
            buf[i] = *dptr ;
            dptr = ((double *)((char *) dptr + stride)) ;
         }
         if (DBWrite(file, const_cast<char *> (name),
                     buf, &length, 1, DB_DOUBLE) == -1) {
            setError(VIO_ERR_WRITE) ;
         }
         delete [] buf ;
      }
   }

   void ArrayInt(const char *name, int *data, int length, int stride)
   {
      if (stride == 0 || stride == sizeof(int)) {
         if (DBWrite(file, const_cast<char *> (name),
                     data, &length, 1, DB_INT) == -1) {
            setError(VIO_ERR_WRITE) ;
         }
      }
      else /* strided */ {
         int i ;
         int *buf = new int[length] ;
         int *iptr = data ;
         for (i=0; i<length; ++i) {
            buf[i] = *iptr ;
            iptr = ((int *)((char *) iptr + stride)) ;
         }
         if (DBWrite(file, const_cast<char *> (name),
                     buf, &length, 1, DB_INT) == -1) {
            setError(VIO_ERR_WRITE) ;
         }
         delete [] buf ;
      }
   }

   void String ( const char* name, char* data, int length)
   {
      if (DBWrite(file, const_cast<char *> (name),
                  data, &length, 1, DB_CHAR) == -1) {
         setError(VIO_ERR_WRITE) ;
      }
   }

   void ExitScope(void)
   {
      if (DBSetDir(file, const_cast<char *> ("..")) == -1)
         setError(VIO_ERR_SETSCOPE) ;
   }
} ;

