/************************************************************************
 * File    : TableBase.c
 * Purpose : A table is a container for items.  Attribute and alias
 *           features are controlled through tables.
 *
 * Note    : This will PROBABLY end up being a base class for derived
 *           classes such as ParamTable, FieldTable, RelationTable,
 *           and ViewTable.
 ************************************************************************/

#include "Vista.h"
#include "TableBase.h"
#include "VHashTable.h"
#include "Group.h"
#include "Attr.h"
#include "View.h"

/* this is needed for breakpoint strcmp */
#include <string.h>
#include <ctype.h>

static Group *expr(TableBase *tb, const char **pname) ;

/*********************************
 *   Constructors/Destructors 
 ********************************/

TableBase::TableBase(View *enclosingView, TableBase *parentTable, MetaType m)
{
   m_itemTable = VHashTableNew(VBIG_HASH_SIZE) ; /* Create key/value Table */
   m_attrTable = VHashTableNew(VSMALL_HASH_SIZE); /* Create attribute Table */
   m_enclosingView = enclosingView ;
   m_parentTable = parentTable ;
   m_tableType = m ;

   return ;
}

TableBase::~TableBase()
{
   /* delete all keys.  Delete values not of Unknown type. */
   VHashTableFree(&m_itemTable) ;
   VHashTableFree(&m_attrTable) ;

   return ;
}

/*********************************
 *   Item/Table related
 ********************************/

/*
   return an item from the table.
*/

KeyBase *TableBase::get(const char *name)
{
   KeyBase *retVal ;
   View *baseView ;
   const char *baseName ;

   /* if name is a path, get the actual view and name */
   VisPath(m_enclosingView, name, &baseView, &baseName) ;

   switch (m_tableType) {
      case ViewItem :
         if (*baseName == '\0')
            retVal = baseView ;
         else
            retVal = VHashTableFind(baseView->views()->m_itemTable, baseName) ;
         break ;

      case FieldItem :
         retVal = VHashTableFind(baseView->fields()->m_itemTable, baseName) ;
         break ;

      case ParamItem :
         retVal = VHashTableFind(baseView->params()->m_itemTable, baseName) ;
         break ;

      case RelationItem :
         retVal = VHashTableFind(baseView->relations()->m_itemTable, baseName);
         break ;

      default:
         VErrFatal("TableBase::get() -- error in lookup", VERR_INTERNAL) ;
         retVal = 0 ;  /* hard crash */
   }

   return retVal ;

}

/************************************************************************
 * Method    : add()
 * Author(s) : Keasler
 * Purpose   : Do whatever is necessary to add Item to Table
 ************************************************************************/

void TableBase::add(KeyBase *item)
{
   if (!item->isNil()) {
      VHashTableInsert(m_itemTable, item) ;
      item->m_enclosingTable = this ;

      if (item->m_attr != NIL) {
         for (Group::Iter ai(GroupClone(item->m_attr)); !ai.end(); ai.next())
            setAttrGroup(item, ai.item()->name()) ;
      }
      setAttrGroup(item, "_all") ;
   }

   // return item ;
}

/************************************************************************
 * Method    : remove()
 * Author(s) : Keasler
 * Purpose   : Do what's necessary to remove Item from Table.
 *             return removed Item.
 ************************************************************************/

void TableBase::remove( const char *name )
{
   /* Do not change the ValueBase reference count here */

   /* In theory, name does not have to be local, so a Vispath */
   /* operation is needed to be safe... */

   KeyBase *detachedItem = VHashTableRemoveEntry(m_itemTable, name) ;

   if (!detachedItem->isNil()) {
      delAttrGroup(detachedItem, "_all") ;
      detachedItem->m_enclosingTable = NIL ;

      if (detachedItem->m_attr != NIL) {
         for (Group::Iter ai(GroupClone(detachedItem->m_attr));
              !ai.end(); ai.next()) {
            delAttrGroup(detachedItem, ai.item()->name()) ;
         }
      }
   }

   // return detachedItem ;
}



/*
 *   Deletes an item from the table 
 */
void TableBase::del(const char *name)
{
   /* when I have time... */
   name = name ;
}

/*********************************
 *   Attribute related
 ********************************/

static const char *lexan(const char **pstr) {
   static char buf[80] ;

   int i = 0 ;
   char const *str = *pstr ;
   if (*str != 0) {
      if (isalpha(*str) || *str == '_') {
         while (isalpha(*str) || *str == '_')
            buf[i++] = *str++ ;
      }
      else
         buf[i++] = *str++ ;
   }
   buf[i] = 0 ;

   while (*str != 0 && isspace(*str))
      ++str ;

   *pstr = str ;

   return buf ;
}

static Group *factor(TableBase *tb, const char **pname)
{
   Group *retGroup ;

   if (**pname == '(') {
      lexan(pname) ;
      retGroup = expr(tb, pname) ;
      if (**pname == ')')
         lexan(pname) ;
   }
   else {
      retGroup = (Group *) VHashTableFind(tb->m_attrTable, lexan(pname)) ;
      if (retGroup->isNil()) {
         /* If they ask for a non-existent attribute, return NIL. */
         /* Note that if they are using the group inverse (~), then */
         /* a NIL group might turn into a valid expression. */
         /* We should really return a NIL singleton here rather than */
         /* allocating a new NIL group.  The Group deletion would */
         /* just need to have an adequate guard, in that case. */
         /* That leads to complications though, because the user */
         /* wants to be able to use the delete operation on this object */
         /* Maybe the best thing to do is to add a NIL, locked, and */
         /* deletable attribute to each item. */
         retGroup = new Group("Nil", 0) ;
         retGroup->m_enclosingTable = tb ;
      }
      else
         retGroup = GroupClone(retGroup) ;
   }

   return retGroup ;
}

static Group *term(TableBase *tb, const char **pname)
{
   Group *retGroup ;
   int parity = 0 ;
   while (**pname == '~') {
     lexan(pname) ; /* discard '~' */
     parity = 1 - parity ;
   }
   if (parity) {
      Group *right = factor(tb, pname) ;
      retGroup = GroupInverse(0, right) ;
      delete right ;
   }
   else
      retGroup = factor(tb, pname) ;

   return retGroup ;
}

static Group *expr(TableBase *tb, const char **pname)
{
   Group *retGroup ;
   Group *left, *right ;
   Group *(*func)(const char *, Group*, Group*) ;

   retGroup = term(tb, pname) ;
   while(1) {
      switch (**pname) {
         case '&': func = GroupIntersect ; break ;
         case '|': func = GroupUnion     ; break ;
         case '-': func = GroupSubtract  ; break ;
         case '^': func = GroupDiff      ; break ;
          default: goto end_expr ;
      }

      lexan(pname) ; /* consume binary operator token */
      left = retGroup ;
      right = term(tb, pname) ;
      retGroup = func(0, left, right) ;
      delete left ;
      delete right ;
   }
end_expr:
   return retGroup ;
}

Group* TableBase::attr(const char *name, int depth)
{
   Group *retVal ;
   const char *tmpName, *tmpNameCooked ;

   if (name == NIL || *name == 0)
      name = "_all" ;

   tmpName = name ;
   while (*tmpName != 0 && isspace(*tmpName))
      ++tmpName ;
   tmpNameCooked = tmpName ;

   retVal = expr(this, &tmpName) ;

   if (depth != 0) {
      TableBase *currTable = this ;
      if (depth < 0) /* search ancestors */ {
         for (int i=depth ; i != 0; ++i) {
            Group *tmp, *retVal2 ;

            currTable = currTable->m_parentTable ;
            if (currTable != NIL) {
               tmpName = tmpNameCooked ;
               retVal2 = expr(currTable, &tmpName) ;
               /* carefully propagate "Nil" name? */
               tmp = GroupUnion(0, retVal, retVal2) ;
               delete retVal ;
               delete retVal2 ;
               retVal = tmp ;
            }
            else
               break ;
         }
      }
      else /* search children */ {
         Group *tmp, *retVal2 ;
         for (Group::Iter vi(m_enclosingView->views()); !vi.end(); vi.next()) {
            View *currView = (View *) vi.item() ;
            switch (m_tableType) {
               case ViewItem:
                  retVal2 = currView->views()->attr(tmpNameCooked, depth-1) ;
                  break ;

               case FieldItem:
                  retVal2 = currView->fields()->attr(tmpNameCooked, depth-1) ;
                  break ;

               case ParamItem:
                  retVal2 = currView->params()->attr(tmpNameCooked, depth-1) ;
                  break ;

               case RelationItem:
                  retVal2 = currView->relations()->attr(tmpNameCooked, depth-1) ;
                  break ;

               default:  /* uh-oh... */ ;
            }
            /* preserve "Nil" name? */
            tmp = GroupUnion(0, retVal, retVal2) ;
            delete retVal ;
            delete retVal2 ;
            retVal = tmp ;
         }
      }
   }

   if (retVal->m_keyName == NIL)
      retVal->m_keyName = VStrdup(name) ;

   return retVal ;
}


/*
 * Add an attribute to item.  There are key and value attributes,
 * so we must distinguish the two.  Add key attribute by default.
 *
 * 'item' is the object the attribute will be added to.
 * 'attrName' is the text string of the attribute
 * 'attrItem' holds any data associated with the attribute
 */
void TableBase::setAttrGroup(KeyBase *item, const char *attrName)
{
   Group *attrGroup = (Group *)VHashTableFind(m_attrTable, attrName) ;
   if (attrGroup->isNil()) {
      attrGroup = new Group(attrName) ;
      attrGroup->m_enclosingTable = this ;
      VHashTableInsert(m_attrTable, attrGroup) ;
   }

   attrGroup->add(item) ;
}


KeyBase *TableBase::setAttr(KeyBase *item,
                             const char *attrName, KeyBase *attrItem)
{
   static KeyBase empty("Nil") ;  /* avoid crashes */

   /* reference counting on attrItem is handled elsewhere. */
   /* Be paranoid if you try to change where the reference counting occurs */

   /* Note that this is optimized for best performance in the general case */
   /* at the sake of readbility. Speed of this operation may be critical. */
   /* BE VERY CAREFUL IF YOU TRY TO MODIFY THIS.  CHANCES ARE THAT YOU */
   /* WILL BE MISSING AN END CASE. */

   if (attrItem == NIL)
      attrItem = &empty ;

   if (item->m_attr != NIL) {
      Attr *oldAttr = (Attr *)item->m_attr->get(attrName) ;
      if (!oldAttr->isNil()) {
         KeyBase *oldAttrItem = oldAttr->value().itemPtr ;
         if (oldAttrItem == attrItem)
            return item ;

         if (oldAttr->m_keyCount == 1) {
            if (oldAttr->m_valuePtr->m_valueCount == 1) {
               if (!oldAttrItem->isNil()) {
                  if (--oldAttrItem->m_keyCount == 0)
                     delete oldAttrItem  ;
               }
               oldAttr->value().itemPtr = attrItem ;
            }
            else {
              --oldAttr->m_valuePtr->m_valueCount ;
              oldAttr->m_valuePtr = (ValueBase *)
                 new AttrPrivate(attrItem, oldAttr) ;
            }
            return item ;
         }
         else {
            --oldAttr->m_keyCount ;
            item->m_attr->del(oldAttr) ; /* shallow delete attr from group */
         }
      }
   }
   else
      item->m_attr = new Group(item->name()) ;

   KeyBase *localAttrItem = new Attr(attrName, attrItem) ;
   item->m_attr->add(localAttrItem) ;
   setAttrGroup(item, attrName) ;

   return item ;
}

/*
 * Remove an attribute to item.  There are key and value attributes,
 * so we must distinguish the two.  Add key attribute by default.
 */
void TableBase::delAttrGroup(KeyBase *item, const char *attrName)
{
   Group *attrGroup = (Group *)VHashTableFind(m_attrTable, attrName) ;
   if (attrGroup->isNil()) {
      char msg[120];
      sprintf(msg,
         "TableBase::delAttrGroup -- Item named %s does not have attribute %s",
             item->name(),  attrName) ;
      VErrFatal(msg, VERR_INTERNAL) ;
   }

   attrGroup->del(item) ;

   return ;
}

void TableBase::delAttr(KeyBase *item, const char *attrName)
{
   KeyBase *attrItem = item->m_attr->get(attrName) ;
   if (!attrItem->isNil()) {
      item->m_attr->del(attrItem) ;
      if (--attrItem->m_keyCount == 0)
         delete attrItem ;

      delAttrGroup(item, attrName) ;
   }

   // return item ;
}

/*********************************
 *   Alias related
 ********************************/

/*
 *  Performs the alias operation, and modifies attributes
 *  appropriately.  Return pointer to new item.
 */
KeyBase *TableBase::alias(const char *newName, KeyBase *oldItem)
{
   KeyBase *retItem = get(newName) ;

   /* clean up any existing alias */

   if (retItem == oldItem)
      return retItem ;

   if (!retItem->isNil()) {

      if (--retItem->m_keyCount == 0)
         delete retItem ;
      else
         remove(newName) ;
   }

   if (!oldItem->isNil()) {
      ValueBase *value = oldItem->m_valuePtr ;

      ++(value->m_valueCount) ; /* make note of new association */

      retItem = new KeyBase(newName, value) ;
      add(retItem) ;
   }
   else
      retItem = oldItem ;

   return retItem ;
}

KeyBase *TableBase::alias(const char *newName, const char *oldName)
{
   KeyBase *retItem = get(oldName) ;
   return alias(newName, retItem) ;
}

/* If the named item exists, rename it */
KeyBase *TableBase::rename(const char *newName, const char *oldName)
{
   KeyBase *widget = get(oldName) ;

   if ((!widget->isNil()) && (widget->m_enclosingTable == this)) {
      remove (widget->name()) ;
      widget->rename(newName) ;
      add(widget) ;
   }

   return widget ;
}

