#ifndef __TABLEBASE_H
#define __TABLEBASE_H

#include "Vista.h"
#include "VHashTable.h"

/* We need some way of keeping track of data types for
   dumping and reading restart files.  Some sort of
   enum might work.  There's probably a better way.
*/

/* This might need to be a base class with a separate derived class */
/* for field, param, relation, view */

///
/// The base class providing hashtable management for KeyBase objects.
/// TableBase contains two hash tables, itemTable and attrTable.  The itemTable
/// manages objects which are similar Item types using the name of the Item
/// as a key.  The attrTable returns a Group given the name of an attribute.
/// This Group contains a GroupPrivate which is an array of pointers which
/// points to values in the itemTable having the queried attribute. This
/// provides a way for TableBase and classes derived from it provide a means
/// to quickly access items either by name or by attribute.
///
class TableBase
{
public:
                /// Construct a Table in the enclosingView whose parentTable
                /// of the same Item type appears in the grandparent View.
                TableBase ( View* enclosingView, TableBase* parentTable,
                            MetaType m );
   virtual     ~TableBase ( );

   ///  Add an Item to the current Table, and attach all attributes.
   void        add ( KeyBase *item );
   ///  Get an item out of the current Table given the name key.
   KeyBase   *get ( const char* name );
   ///  Remove an Item from the current Table detatching all its attributes,
   ///  but don't delete the Item itself.
   void        remove ( const char *name );

public:
   /// Create an alias in this Table to an Item of a differnt name in this
   /// Table or another Table.
   KeyBase   *alias ( const char *newName, KeyBase *oldItem );
   /// Create an alias in this Table to an item of a different name
   /// already existing in this table.
   KeyBase   *alias ( const char *newName, const char *oldName );
   /// Get all Items in this table having the given attribute mask.  If
   /// no mask is given, all items in the table are returned.  An attribute
   /// mask can be of the general form "(a & b) - c ^ d | ~(e & f)". These
   /// are the operators for set intersection (&), union (|), subtraction (-),
   /// mutual exclusion (^), negation(~), and grouping (). If a depth is
   /// given, then the search will progress through ancestor Views (negative),
   /// or through levels of child Views (positive).
   ///
   Group      *attr ( const char *name = 0, int depth = 0 );
   /// You should use the delete operator directly on the item you
   /// want to delete rather than use this method.  If you delete a
   /// View, then an entire subtree of the database will be deleted.
   void        del ( const char *name );
// void        del ( KeyBase *item );
   /// delete an attribute in an Item and the current table at the same time.
   void        delAttr ( KeyBase *item, const char *attrName);
   /// Change the name of an existing Item in the current Table.
   KeyBase   *rename ( const char *newName, const char *oldName );
   /// Add an attribute to an Item and the current table at the same time.
   KeyBase   *setAttr ( KeyBase *item, const char *attrName
                        , KeyBase *attrItem = NIL
                        );
   ///  Returns the number of entries in the current table.
   int          size( ) ;

   /// Used to delete an Item from the attrTable VHashTable that is
   /// associated with this TableBase.
   void        delAttrGroup(KeyBase *item, const char *attrName) ;
   /// Used to add an Item to the attrTable VHashTable that is
   /// associated with this TableBase.
   void        setAttrGroup(KeyBase *item, const char *attrName) ;

public:
   VHashTable  *m_itemTable;      ///< key/value pairs.
   VHashTable  *m_attrTable;      ///< attributes and their associated
                                  ///< groups of Items.
   View        *m_enclosingView;  ///< The View this TableBase lives in.
   TableBase   *m_parentTable;    ///<  The TableBase of the same kind
                                  ///<  in the grandparent View.
   MetaType     m_tableType;      ///< FieldTable, ViewTable, RelationTable,
                                  ///< or ParamTable.

private:
  // The default constructor and copy operator should not be called.
                TableBase();
   TableBase   &operator=(const TableBase &) ;
};

inline int TableBase::size(void)
{
   return VHashTableNumEntries(m_itemTable) ;
}

#endif /* __TABLEBASE_H */
