/**************************************************************************
  Module:  VHashTable
  Purpose: Hash table abstract data type for Vista
 ***************************************************************************/

#include "Vista.h"
#include "VHashTable.h"

#include <stdlib.h>
#include <string.h>

/* Local prototypes */
static unsigned HashFunction(const char *word, int tableSize) ;

#define Warn
#define Fail
#define MeanSD

static KeyBase empty("Nil") ;

/************************************************************************
 * Function  : HashTableAccessInit
 * Author(s) : RN
 * Purpose   : Initializes structure for marching through entire hash table
 ************************************************************************/

VHashTraverse *VHashTraverseInit(void)
{
   VHashTraverse *axs = (VHashTraverse *)malloc(sizeof *axs) ;
   axs->ndx = -1 ;
   axs->currEntry = NIL ;
   return axs ;
}


/************************************************************************
 * Function  : HashTableAccessFree
 * Author(s) :
 * Purpose   :
 ************************************************************************/

void VHashTraverseFree(VHashTraverse *axs)
{
   free(axs) ;
}

/************************************************************************
 * Function  : HashTableGetNext
 * Author(s) : RN
 * Purpose   :
 ************************************************************************/

bool VHashTraverseGetNext(VHashTable *hashTable, VHashTraverse *axs)
{
   bool hasNext = false ;

   if ((axs->ndx >= 0) &&
       (axs->currEntry == NIL)) {
      /* This traverse object was just involved in a HashTableDeleteEntry()
       * call which was unable to back up past the current bucket */
      if (hashTable->m_hashHead[axs->ndx] != NIL) {
         axs->currEntry = hashTable->m_hashHead[axs->ndx] ;
         hasNext = true ;
      }
      else {
         /* Bump the bucket and keep lookin' */
         while(axs->ndx+1 < hashTable->m_tableSize) {

            axs->ndx++ ;

            if (hashTable->m_hashHead[axs->ndx] != NIL) {
               axs->currEntry = hashTable->m_hashHead[axs->ndx] ;
               hasNext = true ;
               break ;
            }
         }
      }
   }
   else if ((axs->ndx == -1) ||
            (axs->currEntry->m_nextKey == NIL)) {
      /* If the traverse type's ndx is -1, then this is the first
       * time into this function and we need to scan for the first bucket with
       * an item at the head of the linked list.  If it's not -1, then it's
       * pointing to the hash entry from the *last* time it was called, and
       * we need to bump the pointer to the next.  If *that* pointer is NIL,
       * then we also need to scan - else we can just point to the next
       * thing in the linked list and keep the bucket index the same */

      while(axs->ndx+1 < hashTable->m_tableSize) {

         axs->ndx++ ;

         if (hashTable->m_hashHead[axs->ndx] != NIL) {
            axs->currEntry = hashTable->m_hashHead[axs->ndx] ;
            hasNext = true ;
            break ;
         }
      }
   }
   else {
      axs->currEntry = axs->currEntry->m_nextKey ;
      hasNext = true ;
   }

   return hasNext ;
}


/************************************************************************
 * Function  : HashTraverseEntry
 * Author(s) :
 * Purpose   :
 ************************************************************************/

KeyBase *VHashTraverseEntry(VHashTraverse *axs)
{
   return axs->currEntry ;
}


/************************************************************************
 * Function  : HashFunction()
 * Author(s) : Kehrnigan and Richie
 * Purpose   : Given a word, returns the integer value associated with
 *             that word.
 ************************************************************************/

static unsigned HashFunction(const char *word, int tableSize)
{
   unsigned hashVal ;
   const char *s = word ;
 
   /* Algorithm stolen from K&R C, Second edition, p. 144 */
   for(hashVal = 0 ; *s != '\0' ; s++)
      hashVal = *s + 31 * hashVal ;

   return (hashVal % tableSize) ;
}






/************************************************************************
 * Function  : HashTableNew()
 * Author(s) : Dan Malmer, Rob Neely
 * Purpose   : creates a new hash table
 ************************************************************************/

VHashTable *VHashTableNew(VHashTableSizes_t tableSize)
{
   int i ;
   VHashTable *newTable ;

   newTable = (VHashTable *)malloc(sizeof(*newTable)) ;
   newTable->m_tableSize = tableSize ;
   newTable->m_hashHead = new KeyBase *[tableSize] ;
  
   for(i=0 ; i<tableSize ; i++) 
      newTable->m_hashHead[i] = NIL ;

   newTable->m_totalHashed = 0 ;
   return (VHashTable *)newTable ;
}


/************************************************************************
 * Function  : HashTableInsertEntry
 * Author(s) : Rob Neely (modified by Keasler)
 * Purpose   : Insert an existing Item into a hash table.
 ************************************************************************/

void VHashTableInsert(VHashTable *hashTable, KeyBase *item)
{
   unsigned hashVal ;
   KeyBase *currKey ;

   hashVal = HashFunction(item->name(), hashTable->m_tableSize) ;

   if (hashTable->m_hashHead[hashVal] == NIL) {
      hashTable->m_hashHead[hashVal] = item ;
      hashTable->m_hashHead[hashVal]->m_nextKey = NIL ;
   }
   else {
      /* Add new entries to the end of the list.  this
         eliminates some annoying restart diffs that occurred before
         when new nodesets were added to the beginning of the list. */
      currKey = hashTable->m_hashHead[hashVal] ;
      while (currKey->m_nextKey != NIL)
         currKey = currKey->m_nextKey ;
      
      currKey->m_nextKey = item ;
      item->m_nextKey = NIL ;   /* just to be safe... */
   }

   hashTable->m_totalHashed++ ;

   return ;
}



/************************************************************************
 * Function  : HashTableFind
 * Author(s) : Rob Neely
 * Purpose   : 
 ************************************************************************/

KeyBase *VHashTableFind(VHashTable *hashTable, const char *str)
{
   unsigned hashVal ;
   KeyBase *currKey ;
  
   hashVal = HashFunction(str, hashTable->m_tableSize) ;
  
   currKey = hashTable->m_hashHead[hashVal] ;
   while (currKey != NIL) {
      if (Streq(str, currKey->name()))
         break ;
      
      currKey = currKey->m_nextKey ;
   }
   if (currKey == NIL)
      currKey = &empty ;

   return currKey ;
}


/************************************************************************
 * Function  : HashTableDeleteEntry
 * Author(s) : Rob Neely (heavily modified by Keasler)
 * Purpose   : Deletes the entry with the given tag from the hash
 *             table.  If you're calling this function from within
 *             a HashTraverse() construct, then you need to pass
 *             in the current HashTraverse_t object, else just
 *             pass NIL.
 ************************************************************************/

KeyBase *VHashTableRemoveEntry(VHashTable *table, const char *str, bool kill)
{
   unsigned hashVal ;

   hashVal = HashFunction(str, table->m_tableSize) ;

   /* If there's only one entry in this bucket, then it's a
    * piece of cake */
   if (table->m_hashHead[hashVal] != NIL) {
      KeyBase *currKey ;
      KeyBase *prev = NIL ;
      bool found = false ;

      if (table->m_hashHead[hashVal]->m_nextKey == NIL) {

         currKey = table->m_hashHead[hashVal] ;
         if (!strcmp(str, currKey->name())) {
            table->m_hashHead[hashVal] = NIL ;
            if (kill) {
               if (--(currKey->m_keyCount) == 0) {
                  delete currKey ;
                  currKey = NIL ;
               }
            }
            found = true ;
         }
      }
      else {
         currKey = table->m_hashHead[hashVal] ;

         while(currKey != NIL && strcmp(str, currKey->name())) {
            prev = currKey ;
            currKey = currKey->m_nextKey ;
         }

         if (currKey != NIL) {
            if (prev == NIL) {
               table->m_hashHead[hashVal] = currKey->m_nextKey ;
            }
            else {
               prev->m_nextKey = currKey->m_nextKey ;
            }
            if (kill) {
               if (--(currKey->m_keyCount) == 0) {
                  delete currKey ;
                  currKey = NIL ;
               }
            }
            found = true ;
         }
      }

      if (found) {

         /* Set up the hash traverser to point to the right thing */
         table->m_totalHashed--;

         if (currKey != NIL) {
            currKey->m_nextKey = NIL ;
            return currKey ;
         }
      }
   }

   return &empty ;
}




/************************************************************************
 * Function  : HashTableFree
 * Author(s) : Rob Neely
 * Purpose   : Deletes a hashtable, and its' contents.
 ************************************************************************/

void VHashTableFree(VHashTable **tableIn)
{
   int i ;
   VHashTable *table = *tableIn ;

   for(i=0 ; i<table->m_tableSize ; i++) {
      KeyBase *currKey ;
      
      currKey = table->m_hashHead[i] ;
      while (currKey != NIL) {
         KeyBase *tmp = currKey->m_nextKey ;
         if (--(currKey->m_keyCount) == 0) {
            /* speed up wholesale table deletion by setting NIL here */
            currKey->m_enclosingTable = NIL ;
            delete currKey ;
         }
         currKey = tmp ;
      }
   }

   delete [] (table->m_hashHead) ;
   free(*tableIn) ;
   *tableIn = NIL ;
}


/************************************************************************
 * Function  : HashTableFreeBase
 * Author(s) : Keasler
 * Purpose   : Deletes the base hashtable, but not the entries.
 ************************************************************************/

void VHashTableFreeBase(VHashTable **tableIn)
{
   delete [] ((*tableIn)->m_hashHead) ;
   free(*tableIn) ;
   *tableIn = NIL ;

   return ;
}



/************************************************************************
 * Function  : HashTableNumEntries
 * Author(s) : RN
 * Purpose   : 
 ************************************************************************/
int VHashTableNumEntries(VHashTable *hashTable)
{
   return hashTable->m_totalHashed ;
}


