
#include "Group.h"
#include "View.h"
#include "MemManage.h"
#include "VErr.h"

ViewPrivate::ViewPrivate(IndexSet *indexSet, View *creatorView) :
             ValueBase(creatorView)
{
   TableBase *ct ;
   View *pv = NIL ;

   /* finish ValueBase init */
   m_valueType = ViewItem ;

   if (indexSet == NIL) {
      passthrough = true ;
      indexSet = new IndexSet(0) ;
   }
   else
      passthrough = false ;

   /* Initialize ViewPrivate */
   indexSetPrivate = indexSet ;

   /* get the parent View for this newly created View Item */
   ct = creatorView->m_enclosingTable ;
   if (ct != NIL)
      pv = ct->m_enclosingView ;

   if (pv == NIL) {
      viewTable = new ViewTable(creatorView, NIL) ;
      paramTable = new ParamTable(creatorView, NIL) ;
      fieldTable = new FieldTable(creatorView, NIL) ;
      relationTable = new RelationTable(creatorView, NIL) ;
      level = 0 ;
      rootView = creatorView ;
   } else {
      /* allow each table to directly connect to it's parent table */
      viewTable = new ViewTable(creatorView, pv->value().viewTable) ;
      paramTable = new ParamTable(creatorView, pv->value().paramTable) ;
      fieldTable = new FieldTable(creatorView, pv->value().fieldTable) ;
      relationTable = new RelationTable(creatorView,
                                        pv->value().relationTable) ;
      level = pv->value().level + 1 ;
      rootView = pv->value().rootView ;
   }

   /* Don't have time to implement these right now... */
   signature = 0 ;
   chain = NIL ;

#if DEBUG_LAYER
   numbp = 0 ;
#endif
}

/************************************************************************
 * Method    : length()
 * Author(s) : Keasler
 * Purpose   : Return length scope of specified View or field
 ************************************************************************/

int View::length(const char *path)
{
   View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (*suffix != '\0')
      outView = outView->view(suffix) ; /* assume suffix is a view */

   return outView->indexSet()->length() ;
}

/************************************************************************
 * Method    : map()
 * Author(s) : Keasler
 * Purpose   : Return indexset for specified view or field
 ************************************************************************/


int *View::map(const char *path)
{
   View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (*suffix != '\0')
      outView = outView->view(suffix) ; /* assume suffix is a view */

   return outView->indexSet()->map() ;
}

/************************************************************************
 * Method    : VewRelabel()
 * Author(s) : Keasler
 * Purpose   : Change rootView and View level throughout the View tree.
 ************************************************************************/

static void ViewRelabel(View *my, View *root, int lev)
{
   my->value().rootView = root ;
   my->value().level = lev ;

   for (Group::Iter vi(my->views()); !vi.end(); vi.next())
      ViewRelabel((View *)vi.item(), root, lev+1) ;

   return ;
}

/************************************************************************
 * Method    : attach()
 * Author(s) : Keasler
 * Purpose   : Attach a subview to the current View.  Note that the
 *             Indexset is not relabeled when attached.  MAKE SURE
 *             THE INDEXSET IS MADE COMPATIBLE IF YOU USE THIS OPERATION.
 ************************************************************************/

View *View::attach (View *vv)
{
   if (vv != NIL) {
      if ((!vv->isNil()) && (vv->m_enclosingTable == NIL)) {  /* ok to attach */
         TableBase *pt = this->views() ;

         /* For each of the tables in the view, set the parentTable */
         vv->views()->m_parentTable = this->views() ;
         vv->params()->m_parentTable = this->params() ;
         vv->fields()->m_parentTable = this->fields() ;
         vv->relations()->m_parentTable = this->relations() ;

         /* Relabel the rootview and level throughout the subtree */
         ViewRelabel(vv, this->value().rootView, this->value().level+1) ;

         /* Make this View an official part of the parent View Tree */
         pt->add(vv) ;

         return vv ;
      }
      else if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;

         if (vv->isNil()) {
            const char *status ;
            if (vv->name() != NIL && strcmp(vv->name(), "Nil") == 0)
               status = "Nil" ;
            else
               status = "locked" ;
            sprintf(msg, "View %s attachment blocked, %s View",
                    vv->name(), status) ;
         }
         else /* pt == Nil */ {
            sprintf(msg, "View %s is still attached to problem %s",
                    vv->name(), vv->value().rootView->name()) ;
         }
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
   }

   return view("Nil") ;
}

/************************************************************************
 * Method    : detach()
 * Author(s) : Keasler
 * Purpose   : Attach a subview to the current View.  Note that the
 *             Indexset is not relabeled upon detachment.
 ************************************************************************/

View *View::detach (View *ve)
{  TableBase *pt = ve->m_enclosingTable ;
   if ((!ve->isNil()) && (pt->m_enclosingView == this) ) {

      /* Remove the entry from the View Table */
      pt->remove(ve->name()) ;

      /* For each of the tables in the view, set the parentTable to NIL */
      ve->views()->m_parentTable = NIL ;
      ve->params()->m_parentTable = NIL ;
      ve->fields()->m_parentTable = NIL ;
      ve->relations()->m_parentTable = NIL ;

      /* Relabel the rootview and level throughout the tree */
      ViewRelabel(ve, ve, 0) ;
   }
   else {
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (ve->name() != NIL && strcmp(ve->name(), "Nil") == 0)
            status = "Nil" ;
         else if (pt->m_enclosingView != this)
            status = "not a child" ;
         else
            status = "locked" ;
         sprintf(msg, "View %s detachment blocked, %s View",
                 ve->name(), status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
   }

   return ve ;
}

View *View::detach (const char *name)
{  View *ve = view(name) ;
   return detach(ve) ;
}

/************************************************************************
 * Method    : remap
 * Author(s) : Keasler
 * Purpose   : Change the indexset for a View.  Note that all
 *             subcomponents of the view (including subview) are also
 *             remapped/resized/reordered as necessary.
 ************************************************************************/

View *View::remap(IndexSet *is)
{
   int i ;
   int *oldMap = this->map() ;
   int oldLen = this->length() ;
   int *newMap = is->map() ;
   int newLen = is->length() ;
   int minLen = MIN(oldLen, newLen) ;

   /* check to see if the maps are basically the same */
   for (i=0; i<minLen; ++i)
      if (oldMap[i] != newMap[i])
         break ;

   /* The maps are identical, but one is longer than the other */
   if (i == minLen) {
      /* make sure we have work to do... */
      if (oldLen != newLen) {
         /* adjust size of fields in current View. */
         for (Group::Iter fi(fields()); !fi.end(); fi.next()) {
            Field *fd = (Field *) fi.item() ;
            fd->resize(newLen) ;  /* change the length */

            if (newLen > oldLen) {
               if (fd->value().data() == NIL)
                  continue ;

               Param *initVal = (Param *) fd->getAttrItem("init") ;

               /* a check for type() == OTC_UNKNOWN would be nicer... */
               if (!initVal->isNil()) {
                  /* fill uninitialized entries with 'init' value */
                  switch (initVal->type())  {
                     case OTC_INT :
                     {
                        switch(fd->type()) {
                           case OTC_INT:
                           {
                              int val = initVal->Int() ;
                              int *data = fd->Int() ;
                                 for (i=oldLen; i< newLen; ++i)
                                    data[i] = val ;
                           }
                           break ;

                           case OTC_REAL:
                           {
                              real8 val = (real8) initVal->Int() ;
                              real8 *data = fd->Real() ;
                                 for (i=oldLen; i< newLen; ++i)
                                    data[i] = val ;
#if DEBUG
                              if (VErrGetMask() & VERR_CONVERSION) {
                                 char msg[80] ;
                                 sprintf(msg,
                                        "Field is double, fill type is int") ;
                                 VErrWarn(msg, VERR_CONVERSION) ;
                              }
#endif
                           }
                           break ;

                           default: break ;
                        }
                     }
                     break ;

                     case OTC_REAL :
                     {
                        switch(fd->type()) {
                           case OTC_INT:
                           {
                              int val = (int) initVal->Real() ;
                              int *data = fd->Int() ;
                              for (i=oldLen; i<newLen; ++i)
                                 data[i] = val ;
#if DEBUG
                              if (VErrGetMask() &
                                  (VERR_TRUNCATION | VERR_CONVERSION)) {
                                 char msg[80] ;
                                 sprintf(msg,
                                        "Field is int, fill type is double") ;
                                 VErrWarn(msg,
                                        (VERR_TRUNCATION | VERR_CONVERSION)) ;
                              }
#endif
                           }
                           break ;

                           case OTC_REAL:
                           {
                              real8 val = initVal->Real() ;
                              real8 *data = fd->Real() ;
                                 for (i=oldLen; i< newLen; ++i)
                                    data[i] = val ;
                           }
                           break ;

                           default: break ;
                        }
                     }
                     break ;

                     default: break ;
                  }
               }
            }
         }
         /* relations... */
      }

      /* replace old IndexSet with new IndexSet */
      delete value().indexSetPrivate ;
      value().indexSetPrivate = is ;
   }
   else {
      int *oldMatch, *newMatch, numMatch ;

      /* Reorganize the current view and all descendant views */
      /* assume new and old maps are sorted 
            NOTE:  adjacent entries may contain same value or
            map may be unsorted due to slide delete
       */
      VIntersectNodes2(oldMap, oldLen, newMap, newLen,
                       &oldMatch, &newMatch, &numMatch) ;

      for (Group::Iter fi(fields()); !fi.end(); fi.next()) {
         Field *fd = (Field *) fi.item() ;
         if (fd->value().data() == NIL) {
            /* WARNING!!! we actually have to remap every depth */
            /* bucket in this field!!!!! Not done yet... */
            /* Note that this is the Field len, not the memblock len... */
            fd->value().len = newLen ;
            continue ;
         }

#if 0
         /* append new */
         fd->addDepth(1) ;
#endif

         switch(fd->type()) {
            case OTC_INT :
            {
               int *olddata = fd->Int() ;
               int *newdata ;
               memAllocInts(newLen, &newdata, NULL) ;

               /* initialize new */
               /* a check for type() == OTC_UNKNOWN would be nicer... */
               Param *initVal = (Param *) fd->getAttrItem("init") ;
               if (!initVal->isNil()) {
                  switch (initVal->type())  {
                     case OTC_INT :
                     {
                        int val = initVal->Int() ;
                        for (i=0; i<newLen; ++i)
                           newdata[i] = val ;
                     }
                     break ;

                     case OTC_REAL :
                     {
                        int val = (int) initVal->Real() ;
                        for (i=0; i<newLen; ++i)
                           newdata[i] = val ;
#if DEBUG
                        if (VErrGetMask() &
                            (VERR_TRUNCATION | VERR_CONVERSION)) {
                           char msg[80] ;
                           sprintf(msg, "Field is int, fill type is double") ;
                           VErrWarn(msg, (VERR_TRUNCATION | VERR_CONVERSION)) ;
                        }
#endif
                     }
                     break ;

                     default: break ;
                  }
               }
          
               /* copy old to new */
               for (i=0; i<numMatch; ++i)
                  newdata[newMatch[i]] = olddata[oldMatch[i]] ;

               /* release old data and replace with new */
               fd->release() ;
               fd->value().len = newLen ;
               fd->makeSpace((void *)newdata) ;
            }
            break ;

            case OTC_REAL :
            {
               real8 *olddata = fd->Real() ;
               real8 *newdata ;
               memAllocDoubles(newLen, &newdata, NULL) ;

               /* initialize new */
               /* a check for type() == OTC_UNKNOWN would be nicer... */
               Param *initVal = (Param *) fd->getAttrItem("init") ;
               if (!initVal->isNil()) {
                  switch (initVal->type())  {
                     case OTC_INT :
                     {
                        real8 val = (real8) initVal->Int() ;
                        for (i=0; i< newLen; ++i)
                           newdata[i] = val ;
#if DEBUG
                        if (VErrGetMask() & VERR_CONVERSION) {
                           char msg[80] ;
                           sprintf(msg, "Field is double, fill type is int") ;
                           VErrWarn(msg, VERR_CONVERSION) ;
                        }
#endif
                     }
                     break ;

                     case OTC_REAL :
                     {
                        real8 val = initVal->Real() ;
                        for (i=0; i< newLen; ++i)
                           newdata[i] = val ;
                     }
                     break ;

                     default: break ;
                  }
               }
          
               /* copy old to new */
               for (i=0; i<numMatch; ++i)
                  newdata[newMatch[i]] = olddata[oldMatch[i]] ;

               /* release old data and replace with new */
               fd->release() ;
               fd->value().len = newLen ;
               fd->makeSpace((void *)newdata) ;
            }
            break ;

            default: break ;
         }

#if 0
          /* delete old */
          fd->delDepth(1) ;
#endif

       }

       /* replace old map with new map */
       /* for parent -- TO DO */

       /* remap all the children */
      for (Group::Iter vi(views()); !vi.end(); vi.next()) {
         View *vd = (View *) vi.item() ;
         int *oldParent, *oldChild, numChildMatch ;
         int *finalChildMap ;

         /* Intersect view.map() with oldMatch */
         /* and remap child based on result */
         VIntersectNodes2(oldMatch, numMatch, vd->map(), vd->length(),
                          &oldParent, &oldChild, &numChildMatch) ;

         if (numChildMatch != 0) {
            memAllocInts(numChildMatch, &finalChildMap, NULL) ;

            for (i=0; i<numChildMatch; ++i)
               finalChildMap[i] = oldMatch[oldParent[i]] ;

            vd->remap(new IndexSet(numChildMatch, finalChildMap)) ;

            memFree(&finalChildMap, NULL) ;

            /* renumber child to reflect accurate map */
            int *childMap = vd->map() ;
            for (i=0; i<numChildMatch; ++i)
               childMap[i] = newMatch[oldParent[i]] ;

            memFree(&oldParent, &oldChild, NULL) ;
         }
      }

      if (numMatch != 0)
         memFree(&oldMatch, &newMatch, NULL) ;

      /* replace old IndexSet with new IndexSet */
      delete value().indexSetPrivate ;
      value().indexSetPrivate = is ;
   }

   return this ;
}

View *View::splice(const char *name, IndexSet *is)
{
   int oldViewLength = this->length() ;

   if (is->length() != 0) {
      /* create Indexset for new, larger view */
      int newViewLength = this->length() + is->length() ;
      IndexSet *newIS = new IndexSet(newViewLength) ;

      memcpy(newIS->map(), this->map(), this->length()*sizeof(int)) ;
      memcpy(&newIS->map()[this->length()],
             is->map(), is->length()*sizeof(int)) ;

      this->remap(newIS) ;
   }

   /* create child view */
   View *childView = this->viewCreate(name, is) ;

   /* child must point to proper part of parent map */
   if (childView->length() != 0) {
      int i ;
      int *map = childView->map() ;
      int len  = childView->length() ;

      for (i=0; i<len; ++i)
         map[i] = oldViewLength + i ;
   }

   return this ;
}

#if DEBUG

void View::consistencyCheck()
{
   char msg[120] ;
   if (views()->m_enclosingView != this) {
      sprintf(msg, "View %s: View Table has wrong parent View", name()) ;
      VErrWarn(msg, VERR_INTERNAL) ;
   }
   if (params()->m_enclosingView != this) {
      sprintf(msg, "View %s: Param Table has wrong parent View", name()) ;
      VErrWarn(msg, VERR_INTERNAL) ;
   }
   if (fields()->m_enclosingView != this) {
      sprintf(msg, "View %s: Field Table has wrong parent View", name()) ;
      VErrWarn(msg, VERR_INTERNAL) ;
   }
   if (relations()->m_enclosingView != this) {
      sprintf(msg, "View %s: Relation Table has wrong parent View", name()) ;
      VErrWarn(msg, VERR_INTERNAL) ;
   }
   if (m_enclosingTable != NIL) {
      View *enclosingView = m_enclosingTable->m_enclosingView ;

      if (views()->m_parentTable != enclosingView->views()) {
         sprintf(msg, "View %s: View Table has wrong parent Table", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
      if (params()->m_parentTable != enclosingView->params()) {
         sprintf(msg, "View %s: Param Table has wrong parent Table", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
      if (fields()->m_parentTable != enclosingView->fields()) {
         sprintf(msg, "View %s: Field Table has wrong parent Table", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
      if (relations()->m_parentTable != enclosingView->relations()) {
         sprintf(msg,
                 "View %s: Relation Table has wrong parent Table", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
      if (value().level != enclosingView->value().level + 1) {
         sprintf(msg, "View %s: View level is not parent level + 1", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
      if (value().rootView != enclosingView->value().rootView) {
         sprintf(msg,
                 "View %s: rootView is inconsistent with parent", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }

      if (enclosingView->view(name()) != this) {
         sprintf(msg,
                 "View %s: name not consistent with parent Table", name()) ;
         VErrWarn(msg, VERR_INTERNAL) ;
      }
   }
#if EXTENDED_CHECK
   views()->consistencyCheck() ;
   params()->consistencyCheck() ;
   fields()->consistencyCheck() ;
   relations()->consistencyCheck() ;
#endif

   for (Group::Iter vi(views()); !vi.end(); vi.next()) {
      View *child = (View *) vi.item() ;
      child->consistencyCheck() ;
   }
}

#endif /* DEBUG */
