/**************************************************************************
  Module:  View.h
  Purpose: A View provides a scoping mechanism for entities within a mesh.
           Examples of mesh scopes are: all elements in the mesh, all nodes
           in the mesh, elements of a given material type, etc. Each scope
           has its own namespace, and supports Parameter (non-bulk) data,
           and Field (bulk) data.  A View also contains an indexset which
           indicates which subset of the mesh the View scope applies to.
           Note that Views can be nested inside of other Views, and Views
           at lower levels (children) inherit data from Views at higher
           levels (ancestors).
 ***************************************************************************/


#ifndef __VIEW_H
#define __VIEW_H

#include "Vista.h"
#include "IndexSet.h"
#include "ValueBase.h"
#include "ViewTable.h"
#include "ParamTable.h"
#include "FieldTable.h"
#include "RelationTable.h"
#include "Param.h"
#include "Field.h"
#include "Relation.h"
#include "KeyBase.h"

/* pick up the sprintf() function prototype */
#include <stdio.h>
#include <string.h>

///
/// Data private to a View.
/// The descendent of ValueBase providing data encapsulation specific to class
/// View which is a descendent of the parallel KeyBase hierarchy.
///
class ViewPrivate : public ValueBase {
public:  /* should be protected: when we go to multiple inheritance */
         /* We can't make it protected now becasue this class is accessed */
         /* through a pointer, so the permissions don't work. There's */
         /* no point in using friends becasue it's going to change later. */

   ViewPrivate ( IndexSet* indexSet, View* creatorView );
  ~ViewPrivate ( );

   IndexSet*&       indexSet ( );
   ParamTable*&     params ( );
   FieldTable*&     fields ( );
   RelationTable*&  relations ( );
   ViewTable*&      views ( );
   View*            view ( const char* name );

public:
   /* The dataTypes below have a memory overhead.  At the moment */
   /* they are all allocated at view construction time.  They can */
   /* be changed so that they are allocated only if they are used. */
   /* This would be done by setting them all to NIL pointers at */
   /* construction time, and adding a check in each of the lookup */
   /* methods that will construct them if they are NIL.  */

   /* The reason they are allocated statically now is becasue in the */
   /* grand scheme of things, wasting 16K per mesh view is not going */
   /* to kill anyone.  128MB is currently $40.00, and it will only */
   /* become cheaper and larger/faster in the future. */

   /* Note that the four tables below can be aggregated into one */
   /* big table.  They are kept as four tables below simply to give */
   /* each object its own namespace.  It doesn't cost much to do this */
   /* so it should be ok. */

   IndexSet*      indexSetPrivate ;
   ViewTable*     viewTable ;
   ParamTable*    paramTable ;
   FieldTable*    fieldTable ;
   RelationTable* relationTable ;
   int            signature ;    /* to locate view within prob tree */
   int            level ;        /* depth in the view tree */
   View*          chain ;        /* for chained views */
   View*          rootView ;     /* root of problem tree */
   bool           passthrough ;  ///< Dictates behavior of IndexSet

} ;
   

///
/// Vista Items providing a scoping mechanism for entities within a mesh.
/// Examples of mesh scopes are: all elements in the mesh, all nodes in the
/// mesh, elements of a given material type, etc. Each scope has its own
/// namespace, and supports Param eter (non-bulk) data, and Field (bulk) data.
/// A View also contains an IndexSet which indicates which subset of the mesh
/// the View scope applies to.  Note that Views can be nested inside of other
/// Views, and Views at lower levels (children) inherit data from Views at
/// higher levels (ancestors).
/// This class is derived from KeyBase with many member functions specific
/// to Views.
///
class View : public KeyBase {
public:
   /* fill item fields, construct View */
                   View ( const char* name, IndexSet* indexSet=NIL,
                          TableBase* enclosingTable=NIL );
                   /// Create a View having the given name with the given
                   /// IndexSet, and give the pointer to the ViewTable
                   /// that the View was created in, if applicable.
                   View ( const char* name, int idx, IndexSet* indexSet=NIL,
                          TableBase* enclosingTable=NIL );
                  ~View ( );

   ViewPrivate&    value ( );

   /* core functions -- copies of all methods from View */
   /* If/When we switch to multiple inheritance, these go away */

   View*           view ( const char* name );
   /// Return a subView of the given name.
   View*           view ( const char* name, int idx );

   /// Return the IndexSet containing indicies of the entities in the parent
   /// View that are related to the entities defined in the current View.
   IndexSet*&      indexSet ( );
   ParamTable*&    params ( );
   /// Return the ParamTable associated with this View.
   ParamTable*&    params ( const char* path );
   FieldTable*&    fields ( );
   /// Return the FieldTable associated with this View.
   FieldTable*&    fields ( const char* path );
   RelationTable*& relations ( );
   /// Return the RelationTable associated with this View.
   RelationTable*& relations ( const char* path );
   ViewTable*&     views ( );
   /// Return the ViewTable containing children of this View.
   ViewTable*&     views ( const char* path );

   void            lock ( ) ; /* allow queries, disable create/destroy */
   /// A locking mechanism for Table entries (in development)
   void            unlock ( ) ; /* enable modifications */

   /// Attach a subView to this View.
   View*           attach ( View *view );
   View*           detach ( View *view );
   /// Detach a subView of this View.
   View*           detach ( const char *name );

   /// Remap the current View given the new IndexSet, and then recursively
   /// remap child Views, adding, deleting or rearranging entities as
   /// necessary, and setting initial values if the "init" attribute is
   /// present.
   View*           remap ( IndexSet* is ); /* change view indexset */
   /// Extend the current View by adding a new "layer" of entities such as
   /// ghost zones.
   View*           splice ( const char* name, IndexSet* is ); /* concat views */

#if DEBUG

   ///  This method checks the linkages between this View and its Tables
   ///  and its parent and Children Views.
   void            consistencyCheck() ; /* Scan View tree for problems */

#endif

   /* The functions below are convenience function that allow you
      to use a very short name to invoke an operation that would
      otherwise require a very long name.  Although these extra
      member functions ignore the dictum that an object is supposed to
      have a specific purpose and a small number of interfaces, it's
      a better implementation because the code is easier to use.  If we
      drop the convenience functions, the code would be truly object
      oriented, where each object is orthogonal and does only what
      it needs to.  As implemented, practical considerations overide
      theoretical considerations, as they should.
    */

   View*           viewCreate ( const char *name, IndexSet *is = NIL );
   ///  Create a subView
   View*           viewCreate ( const char *name, int idx, IndexSet *is = NIL );
   ViewTable*      viewDestroy ( const char *name );
   ///  Remove a subView (delete operator is a better way to go)
   ViewTable*      viewDestroy ( const char *name, int idx );

   /* indexSet related */
   int             length ( );
   /// Number of entities in the InsexSet for this View.
   int             length ( const char* path );
   int*            map ( );
   /// Indices of related entities in the parent View.
   int*            map ( const char* path );

   /* param related */
   Param*          param ( const char* name );
   Param*          paramSetInt ( const char* name, int value );
   Param*          paramSetReal ( const char* name, double value );
   Param*          paramSetIntArray ( const char* name, int *value,
                                      int n, bool copyMe = VISTA_OWNS );
   Param*          paramSetRealArray ( const char* name, double *value,
                                       int n, bool copyMe = VISTA_OWNS );
   Param*          paramSetString ( const char* name, const char* value );

   int&            paramInt ( const char* name );
   double&         paramReal ( const char* name );
   char*&          paramString ( const char* name );

   /* field related */
   Field*          field ( const char* name );
   Field*          fieldCreateInt ( const char* name, int depth = 1, int** data = NIL );
   Field*          fieldCreateReal ( const char* name, int depth = 1, double** data = NIL );
   Field*          fieldCreateDynamic ( const char *name,
                                        void (*f)(KeyBase*), int depth = 1 );
   int*&           fieldInt ( const char *name );
   int*&           fieldInt ( const char *name, int n );
   double*&        fieldReal ( const char *name );
   double*&        fieldReal ( const char *name, int n );
   void            fieldRelease ( const char *name );
   void            fieldRelease ( const char *name, int n );

   /* relation related */
   Relation*       relation ( const char *name );
   Relation*       relationCreate ( const char *name, int fixedSize, View *otherView = NIL );

private:
  // The default constructor and copy operator should not be called.
                   View();
   View           &operator=(const View &) ;
} ;





/*********************************
 *        ViewPrivate Internal
 *********************************/

inline ViewPrivate::~ViewPrivate ( )
{  delete viewTable ;
   delete indexSetPrivate ;
   delete paramTable ;
   delete fieldTable ;
   delete relationTable ;

}

inline IndexSet *&ViewPrivate::indexSet ( )
{  return indexSetPrivate;
}

inline ParamTable *&ViewPrivate::params ( )
{  return paramTable;
}

inline FieldTable *&ViewPrivate::fields ( )
{  return fieldTable;
}

inline RelationTable *&ViewPrivate::relations ( )
{  return relationTable;
}

inline ViewTable *&ViewPrivate::views ( )
{  return viewTable;
}

inline View *ViewPrivate::view ( const char* name )
{  return views()->view(name);
}








/*********************************
 *          View "Internal"
 *********************************/

inline View::View ( const char *name, IndexSet *indexSet,
                    TableBase *enclosingTable
                  ) : KeyBase(name, NIL, enclosingTable) 
{  m_valuePtr = (ValueBase*) new ViewPrivate(indexSet, this) ;
}

inline View::View ( const char *name, int idx,
                    IndexSet *indexSet, TableBase *enclosingTable
                  ) : KeyBase(0, NIL, enclosingTable) 
{  char fullName[256] ;
   sprintf(fullName, "%s%d", name, idx) ;
   m_keyName = VStrdup(fullName) ;
   m_valuePtr = (ValueBase*) new ViewPrivate(indexSet, this) ;
}

inline View::~View ( )
{  /* do nothing */ }

inline ViewPrivate& View::value ( )
{  return *((ViewPrivate*)m_valuePtr);
}

inline View *View::view ( const char* name )
{  return value().view(name);
}

inline View *View::view ( const char* name, int idx )
{
   /* Depth should be a part of the Table class, but this is */
   /* a quick hack (but slow) to demonstrate the functionality */

   char fullName[256] ;
   sprintf(fullName, "%s%d", name, idx) ;
   return value().view(fullName) ;
}


inline IndexSet *&View::indexSet ( )
{  return value().indexSet();
}

inline ParamTable *&View::params ( )
{  return value().params();
}

inline ParamTable *&View::params ( const char* path )
{  View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (!outView->isNil())
      outView = outView->view(suffix) ;

   return outView->params() ;
}

inline FieldTable *&View::fields ( )
{  return value().fields();
}

inline FieldTable *&View::fields ( const char* path )
{  View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (!outView->isNil())
      outView = outView->view(suffix) ;

   return outView->fields() ;
}

inline RelationTable *&View::relations ( )
{  return value().relations();
}

inline RelationTable *&View::relations ( const char* path )
{  View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (!outView->isNil())
      outView = outView->view(suffix) ;

   return outView->relations() ;
}

inline ViewTable *&View::views ( )
{  return value().views();
}

inline ViewTable *&View::views ( const char* path )
{  View *outView ;
   const char *suffix ;

   VisPath(this, path, &outView, &suffix) ;
   if (!outView->isNil())
      outView = outView->view(suffix) ;

   return outView->views() ;
}

inline void View::lock ( )
{  m_isNilObject = true ;
}

inline void View::unlock ( )
{  if ((name() == NIL) || (strcmp(name(), "Nil") != 0))
      m_isNilObject = false ;
}

inline View *View::viewCreate ( const char* name, IndexSet* is )
{  View *outView ;
   const char *suffix ;

   VisPath(this, name, &outView, &suffix) ;
   if (!outView->isNil())
      outView = outView->views()->create(suffix, is) ;
#if DEBUG
   else {
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (outView->name() != NIL && 
             strcmp(outView->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "View %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
   }
#endif

   return outView ;
}

inline View *View::viewCreate ( const char* name, int idx, IndexSet* is )
{  char fullName[256] ;
   sprintf(fullName, "%s%d", name, idx) ;
   return viewCreate(fullName, is) ;
}

inline ViewTable *View::viewDestroy ( const char* name )
{  View *discard = view(name) ;
   if (!discard->isNil())
      delete discard ;
#if DEBUG
   else {
      if (VErrGetMask() & (VERR_NOTCREATED | VERR_NOTFOUND)) {
         char msg[120] ;
         if (discard->name() != NIL && 
             strcmp(discard->name(), "Nil") == 0) {
            if (VErrGetMask() & VERR_NOTFOUND) {
               sprintf(msg, "View %s doesn't exist", name) ;
               VErrWarn(msg, VERR_NOTFOUND) ;
            }
         }
         else if (VErrGetMask() & VERR_NOTCREATED) {
            sprintf(msg, "View %s is locked", discard->name()) ;
            VErrWarn(msg, VERR_NOTCREATED) ;
         }
      }
   }
#endif
   
   return views() ;
}

inline ViewTable *View::viewDestroy ( const char *name, int idx )
{  char fullName[256] ;
   sprintf(fullName, "%s%d", name, idx) ;
   return viewDestroy(fullName) ;
}

/*********************************
 *          IndexSet Related
 *********************************/

inline int View::length ( )
{  return indexSet()->length();
}

inline int* View::map ( )
{  return indexSet()->map();
}

/*********************************
 *          Param Related
 *********************************/

inline Param  *View::param ( const char* name )
{  return params()->param(name);
}

inline Param  *View::paramSetInt ( const char* name, int value )
{  if (!isNil())
      return params()->setInt(name, value) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Param %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return params()->param("Nil") ;
   }
}

inline Param  *View::paramSetReal ( const char* name, double value )
{  if (!isNil())
      return params()->setReal(name, value) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Param %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return params()->param("Nil") ;
   }
}

inline Param *View::paramSetIntArray ( const char* name, int *value,
                                        int n, bool copyMe )
{  if (!isNil())
      return params()->setIntArray(name, value, n, copyMe) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Param %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return params()->param("Nil") ;
   }
}

inline Param *View::paramSetRealArray ( const char* name, double *value,
                                         int n, bool copyMe )
{  if (!isNil())
      return params()->setRealArray(name, value, n, copyMe) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Param %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return params()->param("Nil") ;
   }
}

inline Param *View::paramSetString ( const char* name, const char* value )
{  if (!isNil())
      return params()->setString(name, value) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Param %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return params()->param("Nil") ;
   }
}

inline int& View::paramInt ( const char* name )
{  return param(name)->Int();
}

inline double& View::paramReal ( const char* name )
{  return param(name)->Real();
}

inline char*& View::paramString (const char* name )
{  return param(name)->String();
}

/*********************************
 *          Field Related
 *********************************/

inline Field *View::field ( const char* name )
{  return fields()->field(name);
}

inline Field *View::fieldCreateInt ( const char* name, int depth, int** data)
{  if (!isNil())
      return fields()->createInt(name, depth, data);
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Field %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return fields()->field("Nil") ;
   }
}

inline Field *View::fieldCreateReal ( const char* name, int depth, double** data
                                     )
{  if (!isNil())
      return fields()->createReal(name, depth, data) ;
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Field %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return fields()->field("Nil") ;
   }
}

inline Field *View::fieldCreateDynamic ( const char *name, void (*f)(KeyBase*)
                                        , int depth
                                        )
{  if (!isNil())
      return fields()->createDynamic(name, f, depth);
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Field %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return fields()->field("Nil") ;
   }
}

inline int*& View::fieldInt ( const char *name )
{  return field(name)->Int();
}

inline int*& View::fieldInt ( const char *name, int n )
{  return field(name)->Int(n);
} 

inline double*& View::fieldReal ( const char* name )
{  return field(name)->Real();
}

inline double*& View::fieldReal ( const char* name, int n )
{  return field(name)->Real(n);
}

inline void View::fieldRelease ( const char* name )
{  if (!isNil())
      field(name)->release();
#if DEBUG
   else {
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "doesn't exist" ;
         else
            status = "is locked" ;
         sprintf(msg, "Parent view %s, Field %s not released", status, name) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
   }
#endif
}

inline void View::fieldRelease ( const char* name, int n )
{  if (!isNil())
      field(name)->release(n);
#if DEBUG
   else {
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "doesn't exist" ;
         else
            status = "is locked" ;
         sprintf(msg, "Parent view %s, Field %s not released", status, name) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
   }
#endif

}

/*********************************
 *          Relation Related
 *********************************/

inline Relation *View::relation ( const char*name )
{  return relations()->relation(name);
}

inline Relation *View::relationCreate ( const char* name, int fixedSize, View* otherView )
{  if (!isNil())
      return relations()->create(name, fixedSize, otherView);
   else {
#if DEBUG
      if (VErrGetMask() & VERR_NOTCREATED) {
         char msg[120] ;
         const char *status ;
         if (this->name() != NIL && strcmp(this->name(), "Nil") == 0)
            status = "Nil" ;
         else
            status = "locked" ;
         sprintf(msg, "Relation %s creation blocked, %s parent View",
                 name, status) ;
         VErrWarn(msg, VERR_NOTCREATED) ;
      }
#endif
      return relations()->relation("Nil") ;
   }
}

#endif /* __VIEW_H */
