#ifndef __ViewIO_h
#define __ViewIO_h
#include "View.h"

/* Base Class for IO operations
 *
 * This interface is specialized for readers and writers of each device.
 * For example here is how the SetDir method might work for various devices:
 *
 *    File
 *       Writer (IOFileSILOWriter class)
 *          if (DBSetDir(IO.fp(), name) == -1) {
 *             DBMkDir(IO.fp, name) ;
 *             DBSetDir(IO.fp, name) ;
 *          }
 *       Reader (IOFileSILOReader class)
 *          DBSetDir(IO.fp(), name) ;
 *
 *    Memory
 *       Writer (IOMemoryWriter class)
 *          Write a header containing directory name
 *       Reader (IOMemoryReader class)
 *          Read header containing directory name
*/

#define VIO_ERR_FCREATE  1
#define VIO_ERR_FOPEN    2
#define VIO_ERR_MKSCOPE  3
#define VIO_ERR_SETSCOPE 4
#define VIO_ERR_WRITE    5
#define VIO_ERR_READ     6


///
/// The abstract base class for IO operations between Vista and external,
/// persistent formats.
/// This class defines the interface for vista readers and writers.  Derived
/// classes of this abstract base will implement the interface defined here
/// for reading from or writing to specific data formats.
///
class VistaTransfer {
protected:
   int errCode ;
   void setError(int num) {
      errCode = num ; /* set a breakpoint here to find I/O errors */
   }
public:
   VistaTransfer() : errCode(0) { } 
   int error() const { return errCode ; }
   void resetError() {
      errCode = 0 ; /* set breakpoint here for attempted recoveries */
   }
   virtual void EnterScope(const char *name) = 0 ;
   virtual void ArrayReal(const char *name,
                          double *data, int length, int stride = 0) = 0 ;
   virtual void ArrayInt(const char *name,
                         int *data, int length, int stride = 0) = 0 ;
   virtual void String(const char *name, char *data, int length = 0) = 0 ;
   virtual int Length(const char *) { return -1 ; } /* invalid default */
   virtual void ExitScope(void) = 0 ;
   virtual ~VistaTransfer() { }
} ;

void ViewTransfer(VistaTransfer *xfer, View *view, const char *attrMask = NIL) ;
#endif

