
#include "ViewIO.h"
#include "Group.h"
#include "Attr.h"
#include <stdlib.h>

static void WriteAttributes(VistaTransfer *xfer, TableBase *table) ;

static void WriteIndexSet(VistaTransfer *xfer, IndexSet *is)
{
   if (is->length() > 0) {
      if (is->m_dims != 0) {
         int tmp[40] ;
         tmp[0] = -1 ;  // sentinel to indicate this IndexSet is a structured
         tmp[1] = is->m_parentCornerOffset ;
         tmp[2] = is->m_dims ;
         for (int i=0; i < is->m_dims; ++i)
            tmp[3+i] = is->m_extents[i] ;
         for (int i=0; i < is->m_dims; ++i)
            tmp[3 + is->m_dims + i] = is->m_strides[i] ;
         xfer->ArrayInt("Indexset", tmp, 3 + 2*is->m_dims) ;
      }
      else if (is->map() != NIL)
         xfer->ArrayInt("Indexset", is->map(), is->length()) ;
   }
}

static void WriteRelations(VistaTransfer *xfer,
                           RelationTable *relations, const char *attrMask)
{
   // Just query the attribute rather than going through this mess

   if (relations->size() > 0) {
      xfer->EnterScope("Relations") ;
      if (!xfer->error()) {

         for (Group::Iter ri(relations->attr(attrMask)); !ri.end(); ri.next()) {
            Relation *rel = (Relation *) ri.item() ;
            xfer->ArrayInt(rel->name(), rel->value().dataPriv,
                          rel->value().len*rel->value().size) ;
         }

         WriteAttributes(xfer, relations) ;

         xfer->ExitScope() ;
      }
   }
}

static void WriteField(VistaTransfer *xfer, const char *name, Field *fd)
{
   int len = fd->length() ;
   if (len != 0) {
      int stride = fd->value().stride ;

      switch (fd->type()) {
         case OTC_INT:
            xfer->ArrayInt(name, fd->Int(), len, stride) ;
            break ;

         case OTC_REAL:
            xfer->ArrayReal(name, fd->Real(), len, stride) ;
            break ;

         default: /* unknown type */ ;
      }
   }
}

static void WriteFields(VistaTransfer *xfer,
                        FieldTable *fields, const char *attrMask)
{
   // Just query the attribute rather than going through this mess

   if (fields->size() > 0) {
      xfer->EnterScope("Fields") ;
      if (!xfer->error()) {

         for (Group::Iter fi(fields->attr(attrMask)); !fi.end(); fi.next()) {
            Field *fd = (Field *) fi.item() ;
            WriteField(xfer, fd->name(), fd) ;
         }

         WriteAttributes(xfer, fields) ;

         xfer->ExitScope() ;
      }
   }
}

static void WriteParam(VistaTransfer *xfer, const char *name, Param *pd)
{
   switch (pd->type()) {
      case OTC_INT:
         xfer->ArrayInt(name,
                       (int *)pd->value().data(), pd->length()) ;
         break ;

      case OTC_REAL:
         xfer->ArrayReal(name,
                        (double *)pd->value().data(), pd->length()) ;
         break ;

      case OTC_STRING:
      {
         int len = ((pd->length() == 0) ? 3 : (pd->length())) ;
         char *data = ((pd->value().data() == 0) ?
                       ((char *) "#-") : ((char *)pd->value().data())) ;
         xfer->String(name, data, len) ;
      }
      break ;

      default: /* unknown type */ ;
   }
}

static void WriteParams(VistaTransfer *xfer,
                        ParamTable *params, const char *attrMask)
{
   // Just query the attribute rather than going through this mess

   if (params->size() > 0) {
      xfer->EnterScope("Params") ;
      if (!xfer->error()) {

         for (Group::Iter pi(params->attr(attrMask)); !pi.end(); pi.next()) {
            Param *pd = (Param *) pi.item() ;
            WriteParam(xfer, pd->name(), pd) ;
         }

         WriteAttributes(xfer, params) ;

         xfer->ExitScope() ;
      }
   }
}

static void WriteAttributes(VistaTransfer *xfer, TableBase *table)
{
   /* All tables will have the _all so account for that */
   if (VHashTableNumEntries(table->m_attrTable) > 1) {
      int i ;
      char fullname[100] ;
      bool havePublicAttr = false ;

      /* count the number of private attributes (beginning with '_' */
      {
         VHashTraverse *hash = VHashTraverseInit() ;
         while (VHashTraverseGetNext(table->m_attrTable, hash)) {
            Group& ag = (Group &) *VHashTraverseEntry(hash) ;
            if (ag.name()[0] != '_') {
               havePublicAttr = true ;
               break ;
            }
         }
         VHashTraverseFree(hash) ;
      }

      if (havePublicAttr) {
         xfer->EnterScope("Attr") ;
         if (!xfer->error()) {

            /* dump all the attributes in here */
            VHashTraverse *hash = VHashTraverseInit() ;
            while (VHashTraverseGetNext(table->m_attrTable, hash)) {
               Group *ag = (Group *) VHashTraverseEntry(hash) ;

               if (*ag->name() != '_') {
                  int numItem = ag->value().length ;
                  KeyBase **itemPtr = ag->value().itemPtr ;

                  for (i=0; i<numItem; ++i) {
                     KeyBase *it = itemPtr[i] ;
                     sprintf(fullname, "%s_%s", it->name(), ag->name()) ;
                     KeyBase *attrItem = it->getAttrItem(ag->name()) ;
                     if (!attrItem->isNil()) {
                        switch (attrItem->value().m_valueType) {
                           case ParamItem:
                           {
                              Param *pd = (Param *) attrItem ;
                              WriteParam(xfer, fullname, pd) ;
                           }
                           break ;

                           case FieldItem:
                           {
                              Field *fd = (Field *) attrItem ;
                              WriteField(xfer, fullname, fd) ;
                           }
                           break ;

                           default: /* error */ ;
                        }
                     }
                     else {
                        Param attrOnly(fullname, (char *) 0, 0) ;
                        WriteParam(xfer, fullname, &attrOnly) ;
                     }
                  }
               }
            }
            VHashTraverseFree(hash) ;

            xfer->ExitScope() ;
         }
      }
   }

   return ;
}

void ViewTransfer(VistaTransfer* xfer, View *view, const char *attrMask)
{
   char *name = view->name() ; /* get name of this view */

   xfer->EnterScope(name) ;
   if (!xfer->error()) {

      WriteIndexSet(xfer, view->indexSet()) ;
      WriteRelations(xfer, view->relations(), attrMask) ;
      WriteFields(xfer, view->fields(), attrMask) ;
      WriteParams(xfer, view->params(), attrMask) ;

      for (Group::Iter vi(view->views()->attr(attrMask));
                                        !vi.end(); vi.next()) {
         View *child = (View *) vi.item() ;
         ViewTransfer(xfer, child, attrMask) ;
      }

      WriteAttributes(xfer, view->views()) ;

      xfer->ExitScope() ;
   }

   return ;
}

