#include "ViewIO.h"
#include "Group.h"
#include "Attr.h"
#if PARALLEL
#include "mpi.h"
#endif

#if MIN_MAX_NEEDED
#include <ieeefp.h>
#endif

#include <stdio.h>

///
/// A class to implement writing to an external persistent format readable by
/// Visit from Vista.
/// This class implements all the pure virtual functions of class VistaTransfer.
///
class VisitWriter : public VistaTransfer {

   VistaTransfer *lxfer ;

   int idx ;
   int maxIdx ;

   void checkMem() {
      int i ;
      if (idx + 120 > maxIdx) {
         maxIdx *= 2 ;
         char *newbuf = new char[maxIdx] ;
         for (i=0; i<idx; ++i)
            newbuf[i] = buf[i] ;
         delete [] buf ;
         buf = newbuf ;
      }
   }

public:
   char *buf ;
   VisitWriter(VistaTransfer *_lxfer) : lxfer(_lxfer), idx(0), maxIdx(16000) {
      buf = new char[maxIdx] ;
   }
   ~VisitWriter() {
#if PARALLEL
      int i ;
      int nprocs ;
      int myProc ;
      int *scratch, *lengthBuf ;
      int *disp ;
      char *fullString ;
      int totalLength = 0 ;

      MPI_Comm_size(MPI_COMM_WORLD, &nprocs) ;
      MPI_Comm_rank(MPI_COMM_WORLD, &myProc) ;

      scratch = new int[nprocs] ;
      lengthBuf = new int[nprocs] ;
      for (i=0; i<nprocs; ++i)
         scratch[i] = 0 ;
      scratch[myProc] = idx ;
      MPI_Reduce(scratch, lengthBuf, nprocs,
                 MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD) ;

      if (myProc == 0) {
         disp = new int[nprocs] ;
         disp[0] = 0 ;
         totalLength = lengthBuf[0] ;
         for (i=1; i<nprocs; ++i) {
            disp[i] = disp[i-1] + lengthBuf[i-1] ;
            totalLength += lengthBuf[i] ;
         }
         fullString = new char[totalLength+1] ;
      }

      MPI_Gatherv(buf, idx, MPI_CHAR,
                  fullString, lengthBuf, disp, MPI_CHAR,
                  0, MPI_COMM_WORLD) ;

      if (myProc == 0) {
         fullString[totalLength++] = '\0' ;
         lxfer->String("VisIt", fullString, totalLength) ;
         delete [] fullString ;
         delete [] disp ;
         delete [] lengthBuf ;
         delete [] scratch ;
      }
#else
      buf[idx++] = '\0' ;
      lxfer->String("VisIt", buf, idx) ;
#endif
      delete [] buf ;
   }
   void EnterScope(const char *name) {
      checkMem() ;
      while (*name != '\0')
         buf[idx++] = *name++ ;
   }
   void ArrayReal(const char *, double *, int , int ) { }
   void ArrayInt(const char *, int *, int , int ) { }
   void String(const char *name, char *, int) {
      checkMem() ;
      while (*name != '\0')
         buf[idx++] = *name++ ;
   }
   void ExitScope(void) { }
} ;

static void WriteRelations(VistaTransfer *xfer, RelationTable *relations)
{
   // Just query the attribute rather than going through this mess

   if (relations->size() > 0) {
      int pathIdx ;
      char path[200] ;
      char line[300] ;
      xfer->EnterScope("{ ") ;

      for (Group::Iter ri(relations); !ri.end(); ri.next()) {
         Relation *rel = (Relation *) ri.item() ;
         View *other = rel->value().otherView ;

         pathIdx = 198 ;
         while (other->value().rootView != NIL) {
            int len = strlen(other->name()) ;
            pathIdx -= len ;
            memcpy(&path[pathIdx], other->name(), len) ;
            path[--pathIdx] = '/' ;
            if (other->m_enclosingTable == NIL)
               break ;
            other = other->m_enclosingTable->m_enclosingView  ;
         }
         if (pathIdx == 198) {
            path[198] = '/' ;
            path[199] = '\0' ;
         }
         
         sprintf(line, "%s:%d:%s ", rel->name(), rel->value().size,
                 &path[pathIdx]) ;
         xfer->String(line, NIL, 0) ;
      }

      xfer->String("} ", NIL, 0) ;
      xfer->ExitScope() ;
   }
}
static void WriteField(VistaTransfer *xfer, const char *name, Field *fd)
{
   int i ;
   char line[80] ;
   int len = fd->length() ;

   if (len != 0) {
      switch (fd->type()) {
         case OTC_INT:
         {
            int min, max ;
            int *buf = fd->Int() ;
            min = max = buf[0] ;
            for (i=1; i<len; ++i) {
               if (buf[i] < min) min = buf[i] ;
               else if (buf[i] > max) max = buf[i] ;
            }
            sprintf(line, "%s:i:%d:%d ", name, min, max) ;
            xfer->String(line, NIL, 0) ;
            break ;
         }

         case OTC_REAL:
         {
            double min, max ;
            double *buf = fd->Real() ;
            min = max = buf[0] ;
#if MIN_MAX_NEEDED
            for (i=0; i<len; ++i) {
               if (finite(buf[i])) {
                  if (buf[i] < min) min = buf[i] ;
                  else if (buf[i] > max) max = buf[i] ;
               }
            }
#endif
            sprintf(line, "%s:d:%e:%e ", name, min, max) ;
            xfer->String(line, NIL, 0) ;
            break ;
         }

         default: /* unknown type */ ;
      }
   }
}

static void WriteFields(VistaTransfer *xfer, FieldTable *fields)
{
   // Just query the attribute rather than going through this mess

   if (fields->size() > 0) {
      xfer->EnterScope("[ ") ;

      for (Group::Iter fi(fields); !fi.end(); fi.next()) {
         Field *fd = (Field *) fi.item() ;
         WriteField(xfer, fd->name(), fd) ;
      }

      xfer->String("] ", NIL, 0) ;
      xfer->ExitScope() ;
   }
}


static void ViewVisit(VistaTransfer *xfer, View *view)
{
   char line[256] ;
   char *name = view->name() ; /* get name of this view */

   sprintf(line, "( %s:%d ", name, view->length()) ;

   xfer->EnterScope(line) ;

   WriteRelations(xfer, view->relations()) ;
   WriteFields(xfer, view->fields()) ;

   for (Group::Iter vi(view->views()); !vi.end(); vi.next()) {
      View *child = (View *) vi.item() ;
      ViewVisit(xfer, child) ;
   }

   xfer->String(") ", NIL, 0) ;
   xfer->ExitScope() ;

   return ;
}

void ViewVisitDriver(VistaTransfer *xfer, View *view)
{
   extern void VisitParse(char *) ;
   VisitWriter meta(xfer) ;
   ViewVisit(&meta, view) ;
   // VisitParse(meta.buf) ;
   // MPI_Barrier(MPI_COMM_WORLD) ;
}

