
#include "ViewIO.h"
#include <stdio.h>

/***********************************************************************
 * Class  : XMLWriter
 * Author(s) : Keasler
 * Purpose   : Set up a data transfer class for Vista I/O
 **********************************************************************/

///
/// A class to implement writing to the XML external persistent format from
/// Vista.
/// This class implements all the pure virtual functions of class VistaTransfer.
///
class XMLWriter : public VistaTransfer {
   FILE *fp ;
   int indent ;

public:
   XMLWriter(const char *fname, int fileExists = 0) : indent(0) {
      if ((fp = fopen(fname, (fileExists ? "a" : "w"))) == NULL)
         setError((fileExists ? VIO_ERR_FOPEN : VIO_ERR_FCREATE)) ;
      else if (!fileExists) {
         fprintf(fp, "<?xml version='1.0'?>\n") ;
         fprintf(fp, "<!-- VISTA DATABASE -->\n") ;
      }
   }
   ~XMLWriter() { fclose(fp) ; }

   void EnterScope(const char *tag)
   {
      ++indent ;
      if (fprintf(fp, "%*c<View name=\"%s\">\n", 2*indent-1, ' ', tag) < 0)
         setError(VIO_ERR_MKSCOPE) ;
   }

   void ArrayReal(const char *name, double *data, int length, int stride)
   {
      int idx ;
      double *outData = data ;

      if (stride == 0)
         stride = sizeof(double) ;

      for(idx = 0; idx < length ; idx += 20) {
         int bidx ;
         fprintf(fp, "%*c<data name=\"%s\" offset=\"%d\">\n",
                 2*indent+1, ' ', name, idx) ;
         for(bidx = idx; bidx < idx+20 ; bidx += 4) {
            int i ;
            int rowLen = length-bidx ;
            if (rowLen > 4)
               rowLen = 4 ;

            fprintf(fp, "%*c", 2*indent+3, ' ') ;
            for (i=0; i<rowLen; ++i) {
               fprintf(fp, "%15e ", *outData) ;
               outData = (double *) ((char *)outData + stride) ;
            }
            fprintf(fp, "\n") ;
            if (bidx+rowLen == length)
               break ;
         }
         fprintf(fp, "%*c</data>\n", 2*indent+1, ' ') ;
      }
   }

   void ArrayInt(const char *name, int *data, int length, int stride)
   {
      int idx ;
      int *outData = data ;

      if (stride == 0)
         stride = sizeof(int) ;

      for(idx = 0; idx < length ; idx += 50) {
         int bidx ;
         fprintf(fp, "%*c<data name=\"%s\" offset=\"%d\">\n",
                 2*indent+1, ' ', name, idx) ;
         for(bidx = idx; bidx < idx+50 ; bidx += 5) {
            int i ;
            int rowLen = length-bidx ;
            if (rowLen > 5)
               rowLen = 5 ;

            fprintf(fp, "%*c", 2*indent+3, ' ') ;
            for (i=0; i<rowLen; ++i) {
               fprintf(fp, "%8d ", *outData) ;
               outData = (int *) ((char *)outData + stride) ;
            }
            fprintf(fp, "\n") ;
            if (bidx+rowLen == length)
               break ;
         }
         fprintf(fp, "%*c</data>\n", 2*indent+1, ' ') ;
      }
   }

   void String(const char *name, char *data, int length)
   {
      /* this complicated logic is needed for Fortran */

      if (length == 0) {
         fprintf(fp, "%*c<data name=\"%s\" value=\"\"/>\n",
                 2*indent+1, ' ', name) ;
      }
      else {
         int actLen = ((data[length-1] == '\0') ? (length-1) : length) ;
         fprintf(fp, "%*c<data name=\"%s\" value=\"%*.*s\"/>\n",
                 2*indent+1, ' ', name, actLen, actLen, data) ;
      }
   }

   void ExitScope(void)
   {
      if (fprintf(fp, "%*c</View>\n", 2*indent-1, ' ') < 0)
         setError(VIO_ERR_SETSCOPE) ;
      --indent ;
   }

} ;

